#!/usr/bin/env python

# Teuthis 1.0 remote helper application
#
# Paul Ricker
# National Center for Supercomputing Applications
# October 2005
#
# Please see the accompanying README file for the license terms associated
# with your use of this software.

import sys
import os
import os.path
import md5
import binascii

def usage ():
    """
    Print command line syntax.
    """
    print "usage:  tendril.py command args..."

def visit (arg, dirname, names):
    """
    Visit function for os.path.walk which approximates the behavior of os.walk in
    Python 2.4.
    """
    files = [ ]
    dirs  = [ ]
    for item in names:
        if (item == "CVS"):
            names.remove(item)
        else:
            path = os.path.join(dirname, item)
            if not os.path.islink(path):
                if os.path.isdir(path):
                    dirs.append(item)
                else:
                    files.append(item)
    arg.append((dirname, dirs, files))

def md5walk ():
    """
    Walk a directory tree and print filenames and MD5 checksums (sorted).
    """
    path = sys.argv[2]
    tree = [ ]
    os.path.walk(path, visit, tree)
    result = [ ]
    for root, dirs, files in tree:
        for file in files:
            fullname = os.path.join(root, file)
            f        = open(fullname)
            chksum   = md5.new(f.read()).digest()
            f.close()
            normname = normpath(pathlist(fullname))
            result.append((binascii.b2a_hex(chksum), normname, file))
    result.sort()
    for item in result:
        print "%s %s %s" % item

def listfiles ():
    """
    List the files in a path, excluding subdirectories and symbolic links.
    """
    path = sys.argv[2]
    contents = os.listdir(path)
    for entry in contents:
        item = os.path.join(path, entry)
        if os.path.isfile(item) and (not os.path.islink(item)):
            print item
    
def mmkdir ():
    """
    Create directories specified in a file, then remove the file.
    """
    file = sys.argv[2]
    f = open(file)
    data = f.read()
    f.close()
    list = data.split("\n")
    for dir in list:
        path = os.path.normpath(dir)
        if not os.path.exists(path):
            os.makedirs(path)
    os.remove(file)

def normpath (pathlist):
    """
    My normpath takes a path list created by pathlist and returns a Unix-style
    pathname with forward slashes.
    """
    result = ""
    n      = len(pathlist)
    for i in range(n-1):
        result = result + pathlist[i] + "/"
    return result
    
def pathlist (path):
    """
    Given a path string specified in the form native to this machine, return a
    list containing all of the pathname components.  Throws away leading slashes
    (so don't use this on absolute paths).
    """
    dir, name = os.path.split(path)
    if dir and (dir != "/") and (dir != "\\"):
        result = pathlist(dir)
        result.append(name)
        return result
    else:
        return [name]

def main ():
    """
    Main command interpreter.
    """
    if (len(sys.argv) < 2):
        usage()
    else:
        command  = sys.argv[1]
        if (command == "md5walk"):
            md5walk()
        elif (command == "mmkdir"):
            mmkdir()
        elif (command == "listfiles"):
            listfiles()
        else:
            print "tendril:  unknown subcommand"
            sys.exit(1)
 
if __name__ == "__main__":
    main()
