#!/usr/bin/env python

# Teuthis 1.0 simulation management tool
#
# Paul Ricker
# National Center for Supercomputing Applications
# June 2005
#
# Please see the accompanying README file for the license terms associated
# with your use of this software.

import sys, os, os.path

InstallDirectory         = "<<<INSTALLATION PATH>>>"
BinDirectory             = os.path.join(InstallDirectory, "bin")
LibDirectory             = os.path.join(InstallDirectory, "lib")
DocDirectory             = os.path.join(InstallDirectory, "docs")
ShareDirectory           = os.path.join(InstallDirectory, "share")
PixmapDirectory          = os.path.join(ShareDirectory, "pixmaps")
TendrilPath              = os.path.join(BinDirectory, "tendril.py")
GladeFile                = os.path.join(LibDirectory, "teuthis.glade")
HomeDirectory            = os.environ["HOME"]
UserGlobusDirectory      = os.path.join(HomeDirectory, ".globus")

sys.path.append(LibDirectory)

import pygtk
pygtk.require('2.0')
import gtk, gtk.glade, gobject, pango
import time, webbrowser, mailcap, copy
import cPickle, re, pexpect, string, exceptions
import tempfile, shutil, stat
#import select

# Global settings and constants

TempFile                 = tempfile.mktemp()
RemoteScriptFile         = "teuthis.job"

SM_WIKI                  = 0
SM_EDIT                  = 1

SM_FAIL                  = 0
SM_OK                    = 1
SM_CANCEL                = 2
SM_UNKNOWN               = 3
SM_CONTINUE              = 4

SM_TIMEOUT               = 3600

SM_COMPUTE_MACHINE       = 0
SM_ONLINE_DATA_MACHINE   = 1
SM_ARCHIVAL_DATA_MACHINE = 2

SM_RESTART_CONTROL       = 0
SM_RESTART_ARGS          = 1
SM_RESTART_COPY          = 2

SM_STATUS_PEND           = "Pending"
SM_STATUS_RUN            = "Running"
SM_STATUS_DONE           = "Complete"
SM_STATUS_UNK            = "Unknown"
SM_STATUS_NOTAVAIL       = "Unsubmitted"

SM_AUTH_KERBEROS         = 0
SM_AUTH_PASSWORD         = 1
SM_AUTH_HOST             = 2
SM_AUTH_KEYAGENT         = 3
SM_AUTH_GSI              = 4

SM_REMOTE_HOST           = "%h"
SM_REMOTE_HOST2          = "%i"
SM_REMOTE_PATH           = "%r"
SM_REMOTE_USER           = "%u"
SM_REMOTE_USER2          = "%v"
SM_REMOTE_PASS           = "%p"
SM_REMOTE_CMD            = "%c"
SM_REMOTE_TEUTHISDIR     = "%z"
SM_LOCAL_PATH            = "%f"
SM_NUM_CPUS              = "%n"
SM_NUM_NODES             = "%N"
SM_MEM_PER_NODE          = "%m"
SM_WALLTIME              = "%t"
SM_WALLTIME_HOURS        = "%H"
SM_WALLTIME_MINS         = "%M"
SM_WALLTIME_SECS         = "%S"
SM_TILING                = "%T"
SM_QUEUE                 = "%Q"
SM_ACCOUNT               = "%A"
SM_JOB_NAME              = "%L"
SM_BATCH_FILE            = "%b"
SM_REMOTE_JID            = "%j"
SM_PATH_CREATE           = "%D"
SM_CHDIR                 = "%C"
SM_PAR_EXEC              = "%P"
SM_REMOTE_EXEC           = "%e"
SM_ARGS                  = "%a"
SM_REMOTE_RUN_PATH       = "%d"
SM_REMOTE_REALM          = "%R"
SM_CERT_FILE             = "%K"
SM_KEY_FILE              = "%k"
SM_PROXY_LIFE            = "%l"

SM_UPLOAD                = 0
SM_DOWNLOAD              = 1
SM_THIRDPARTY            = 2

SM_OS_WINDOWS            = "Windows"
SM_OS_UNIX               = "Unix"

OSTypes                  = {"Unix":    {"CommandSeparator": "; ", "PathSeparator": "/",  "TestCommand": "pwd", \
                                        "AllFilesWildcard": r'*'}, \
		            "Windows": {"CommandSeparator": "; ", "PathSeparator": "\\", "TestCommand": "dir", \
                                        "AllFilesWildcard": r'*.*'}}

# Exceptions
                                  
class UserCancelEvent(exceptions.Exception):
    def __init__ (self, args=None):
        self.args = args
        
class RemoteExecError(exceptions.Exception):
    def __init__ (self, args=None):
        self.args = args

# GUI class

class AppGUI:
    def __init__(self, store):
        self.Data         = store
        self.wTree        = gtk.glade.XML(GladeFile, "MainWindow")
        self.wTree.window = self.wTree.get_widget("MainWindow")
        self.wTree.splash = self.SplashScreen()
        while gtk.events_pending():
            gtk.main_iteration()
        self.treestore = gtk.TreeStore(str, object, str, str, str, gtk.gdk.Color, \
        			       'gboolean', pango.FontDescription)
        self.treeview  = self.wTree.get_widget("treeview1")
        self.treeview.set_model(self.treestore)
        self.SetTreeViewColumns()
        self.wTree.signal_autoconnect( \
            { "on_project1_activate"      : self.AddProject, \
              "on_open1_activate"         : self.OpenProjectFile, \
              "on_merge1_activate"        : self.MergeProjectFile, \
              "on_save1_activate"         : self.SaveProjectFile, \
              "on_save_as1_activate"      : self.SaveProjectFileAsNew, \
#             "on_export1_activate"       : self.ExportProjectFile, \
#             "on_print1_activate"        : self.PrintProjectFile, \
              "on_update_all_jobs1_activate"  : self.UpdateAllJobStatus, \
              "on_show_description1_activate" : self.ToggleShowDescription, \
              "on_show_status1_activate"  : self.ToggleShowStatus, \
              "on_show_modification_date1_activate" : self.ToggleShowModifiedDate, \
              "on_expand_all1_activate"   : self.ExpandTreeView, \
              "on_collapse_all1_activate" : self.CollapseTreeView, \
              "on_contents1_activate"     : self.HelpWindow, \
              "on_about1_activate"        : self.AboutDialog, \
              "on_treeview1_button_press" : self.TreeViewButtonPress, \
              "on_treeview1_popup_menu"   : self.InvokeOptionsMenu, \
              "on_treeview1_row_activated": self.TreeViewDoubleClick, \
              "on_user_data1_activate"    : self.EditUserData, \
              "on_applications1_activate" : self.EditApplicationData, \
              "on_machines1_activate"     : self.EditMachineData, \
              "gtk_main_quit"             : self.DeleteEvent, \
              "on_close1_activate"        : self.CloseProjectFile, \
              "on_quit1_activate"         : self.Quit, \
              } )
        self.PreferencesDialog = self.CreateDialog("PreferencesDialog", \
            { "entry19"                   : "Name", \
              "entry20"                   : "Organization", \
              "entry21"                   : "Email", \
              "entry40"                   : "Workspace", \
              "colorbutton1"              : "ProjectTextColor", \
              "colorbutton2"              : "ExperimentTextColor", \
              "colorbutton3"              : "RunTextColor", \
              "colorbutton4"              : "JobTextColor", \
              "fontbutton1"               : "ProjectTextFont", \
              "fontbutton2"               : "ExperimentTextFont", \
              "fontbutton3"               : "RunTextFont", \
              "fontbutton4"               : "JobTextFont", \
              "checkbutton13"             : "ConfirmDeletions", \
              "checkbutton14"             : "ShowOutputOnError", \
              "checkbutton15"             : "UseTendril", \
              "checkbutton16"             : "VerboseLogging", \
              "entry101"                  : "LogFile", \
              "entry102"                  : "GridCertDir", \
              "spinbutton17"              : "GridProxyLifetime", \
              "checkbutton17"             : "UseMyProxy", \
              "entry107"                  : "MyProxyServer", \
              "entry109"                  : "MyProxyAccount", \
              "checkbutton18"             : "AutoUpdate", \
              "spinbutton16"              : "AutoUpdateInterval", \
              }, \
            { "button28"                  : ["clicked", self.BrowseForWorkspace], \
              "checkbutton17"             : ["clicked", self.SwitchGridProxyMechanism], \
              } )
        self.ProjectDialog = self.CreateDialog("ProjectDialog", \
            { "entry1"                    : "Name", \
              "textview5"                 : "Comment", \
              "entry81"                   : "NotesLocation", \
              "radiobutton4"              : "NotesType", \
              "label247"                  : "CreatorName", \
              "label6"                    : "BeginDate", \
              "label8"                    : "ModifiedDate", \
              }, \
            { "button36"                  : ["clicked", self.AddExperiment], \
              "button35"                  : ["clicked", self.InvokeNoteTaker], \
              "button37"                  : ["clicked", self.CloneProject], \
              } )
        self.ExperimentDialog = self.CreateDialog("ExperimentDialog", \
            { "entry36"                   : "Name", \
              "textview9"                 : "Comment", \
              "label116"                  : "BeginDate", \
              "label109"                  : "ModifiedDate", \
              "comboboxentry12"           : "Application", \
              "entry87"                   : "ConfigFile", \
              "entry60"                   : "RemoteBuildDir", \
              "entry48"                   : "ConfigCommand", \
              "entry57"                   : "BuildCommand", \
              "entry62"                   : "Executable", \
              "entry58"                   : "ExecArgs", \
              "comboboxentry10"           : "DataSrcMachine", \
              "textview10"                : "DataSrcFiles", \
              "comboboxentry11"           : "DataDestMachine", \
              "entry39"                   : "DataDestPath", \
              "comboboxentry9"            : "ExecMachine", \
              "comboboxentry7"            : "ExecQueue", \
              "comboboxentry8"            : "ExecAccount", \
              "entry72"                   : "NumCPUsRange", \
              "spinbutton11"              : "Tiling", \
              "spinbutton12"              : "MemoryPerNode", \
              "entry41"                   : "ParmFileTemplate", \
              "entry42"                   : "ParmName1", \
              "entry43"                   : "ParmRange1", \
              "checkbutton4"              : "ParmLock1", \
              "entry44"                   : "ParmName2", \
              "entry45"                   : "ParmRange2", \
              "checkbutton5"              : "ParmLock2", \
              "entry46"                   : "ParmName3", \
              "entry47"                   : "ParmRange3", \
              "checkbutton6"              : "ParmLock3", \
              "entry67"                   : "ParmName4", \
              "entry68"                   : "ParmRange4", \
              "checkbutton7"              : "ParmLock4", \
              "entry73"                   : "ParmName5", \
              "entry77"                   : "ParmRange5", \
              "checkbutton9"              : "ParmLock5", \
              "entry74"                   : "ParmName6", \
              "entry78"                   : "ParmRange6", \
              "checkbutton10"             : "ParmLock6", \
              "entry75"                   : "ParmName7", \
              "entry79"                   : "ParmRange7", \
              "checkbutton11"             : "ParmLock7", \
              "entry76"                   : "ParmName8", \
              "entry80"                   : "ParmRange8", \
              "checkbutton12"             : "ParmLock8", \
              }, \
            { "button20"                  : ["clicked", self.GenerateRuns], \
              "button21"                  : ["clicked", self.CloneExperiment], \
              "button32"                  : ["clicked", self.UploadSourceFiles], \
              "button52"                  : ["clicked", self.BrowseForConfigFile], \
              "button53"                  : ["clicked", self.UploadConfigFile], \
              "button33"                  : ["clicked", self.ConfigureApp], \
              "button34"                  : ["clicked", self.BuildApp], \
              "button96"                  : ["clicked", self.ViewConfigOutput], \
              "button97"                  : ["clicked", self.ViewBuildOutput], \
              "button29"                  : ["clicked", self.BrowseForParmTemplate], \
              "button98"                  : ["clicked", self.EditParmTemplate], \
              "comboboxentry12"           : ["changed", self.FillExpExeMachDialogData], \
              "comboboxentry9"            : ["changed", self.FillExpExeMachDialogData], \
              } )
        self.ExperimentDialog.ConfigButton     = self.ExperimentDialog.get_widget("eventbox1")
        self.ExperimentDialog.BuildButton      = self.ExperimentDialog.get_widget("eventbox2")
        self.ExperimentDialog.SavedButtonStyle = self.ExperimentDialog.ConfigButton.get_modifier_style()
        self.RunDialog = self.CreateDialog("RunDialog", \
            { "entry86"                   : "Name", \
              "textview3"                 : "Comment", \
              "comboboxentry23"           : "Disposition", \
              "label234"                  : "BeginDate", \
              "label227"                  : "ModifiedDate", \
              "label281"                  : "Application", \
              "entry84"                   : "Executable", \
              "entry85"                   : "ExecArgs", \
              "label282"                  : "ExecMachine", \
              "comboboxentry19"           : "Queue", \
              "comboboxentry20"           : "Account", \
              "spinbutton13"              : "NumCPUs", \
              "spinbutton14"              : "Tiling", \
              "spinbutton15"              : "MemoryPerNode", \
              "comboboxentry16"           : "DataSrcMachine", \
              "textview4"                 : "DataSrcFiles", \
              "comboboxentry17"           : "DataDestMachine", \
              "entry83"                   : "DataDestPath", \
              "textview1"                 : "ParmVariables", \
              }, \
            { "button41"                  : ["clicked", self.AddJob], \
              "button51"                  : ["clicked", self.AttachRun], \
              "button42"                  : ["clicked", self.CloneRun], \
              "button76"                  : ["clicked", self.ViewRunParameters], \
              } )
        self.JobDialog = self.CreateDialog("JobDialog", \
            { "entry29"                   : "Name", \
              "textview6"                 : "Comment", \
              "comboboxentry22"           : "Disposition", \
              "label82"                   : "BeginDate", \
              "label84"                   : "ModifiedDate", \
              "label283"                  : "Application", \
              "entry69"                   : "Executable", \
              "entry70"                   : "ExecArgs", \
              "label284"                  : "ExecMachine", \
              "comboboxentry2"            : "Queue", \
              "comboboxentry3"            : "Account", \
              "spinbutton7"               : "NumCPUs", \
              "spinbutton8"               : "Tiling", \
              "spinbutton9"               : "MemoryPerNode", \
              "comboboxentry5"            : "DataSrcMachine", \
              "textview7"                 : "DataSrcFiles", \
              "comboboxentry6"            : "DataDestMachine", \
              "entry33"                   : "DataDestPath", \
              "label67"                   : "RemoteJobID", \
              "label68"                   : "SubmitDate", \
              "label69"                   : "RunStatus", \
#             "label75"                   : "DataStatus", \
              "entry89"                   : "WallTimeHrs", \
              "entry90"                   : "WallTimeMin", \
              "entry91"                   : "WallTimeSec", \
              }, \
            { "button13"                  : ["clicked", self.ViewJobParameters], \
              "button14"                  : ["clicked", self.SubmitJob], \
              "button15"                  : ["clicked", self.GetJobStatus], \
              "button18"                  : ["clicked", self.CloneJob], \
              "button48"                  : ["clicked", self.ViewJobLog], \
              "button49"                  : ["clicked", self.ViewJobOutput], \
              "button19"                  : ["clicked", self.ContinueJob], \
              "button87"                  : ["clicked", self.ArchiveJobDataFromButton], \
              } )
        self.ApplicationDialog = self.CreateDialog("ApplicationDialog", \
            { "entry34"                   : "Name", \
              "entry35"                   : "Description", \
              "entry50"                   : "LocalSourcePath", \
              "entry51"                   : "ExecutableName", \
              "entry52"                   : "ConfigFileName", \
              "entry55"                   : "ConfigCommand", \
              "entry56"                   : "BuildCommand", \
              "entry53"                   : "ParamFileName", \
              "entry54"                   : "LogFileName", \
              "checkbutton2"              : "BuildExecutable", \
              "checkbutton3"              : "MoveExecutable", \
              "checkbutton8"              : "IsParallel", \
              "radiobutton6"              : "RestartType", \
              "entry98"                   : "RestartControlFile", \
              "entry99"                   : "RestartArgs", \
              "entry100"                  : "RestartCopyFromFile", \
              }, \
            { "button30"                  : ["clicked", self.BrowseForAppSourceDir], \
              "button57"                  : ["clicked", self.RemoveApplication], \
              "button58"                  : ["clicked", self.NewApplication], \
              "button64"                  : ["clicked", self.CloneApplication], \
#             "treeview3"                 : ["button-press-event", self.ResListTreeViewButtonPress], \
#            "treeview3"                 : ["popup-menu", self.SelectResource], \
              "treeview3"                 : ["changed", self.SelectResource, "get_selection"]
              }, \
            "treeview3" )
        self.PopulateResListTreeView(self.ApplicationDialog, self.Data.Applications)
        self.MachineDialog = self.CreateDialog("MachineDialog", \
            { "entry16"                   : "Name", \
              "textview8"                 : "Description", \
              "radiobutton1"              : "Type", \
              "entry18"                   : "LoginHost", \
              "combobox3"                 : "OSType", \
              "combobox1"                 : "AccessMethod", \
              "entry92"                   : "Realm", \
              "entry111"                  : "GridCertSubject", \
              "combobox2"                 : "QueueSystem", \
              "entry97"                   : "AuthInitCommand", \
              "entry24"                   : "RemoteExecCommand", \
              "entry25"                   : "FileUploadCommand", \
              "entry28"                   : "FileDownloadCommand", \
              "entry112"                  : "ThirdPartyTransferCmd", \
              "entry22"                   : "JobTemplateFile", \
              "spinbutton3"               : "CPUsPerNode", \
              "spinbutton4"               : "DefaultCPUsPerNode", \
              "spinbutton5"               : "MemoryPerNode", \
              "entry31"                   : "RemoteUserid", \
              "comboboxentry24"           : "DefaultQueue", \
              "comboboxentry4"            : "DefaultAccount", \
              "entry59"                   : "RemoteBuildRootDir", \
              "entry61"                   : "RemoteExecDir", \
              "entry88"                   : "RemoteRunRootDir", \
              "entry104"                  : "TeuthisDir", \
#              "entry63"                   : "FileCopyCommand", \
#              "entry64"                   : "FileMoveCommand", \
#              "entry65"                   : "PathCreateCommand", \
#              "entry66"                   : "ChangeDirCommand", \
              "entry71"                   : "ParallelExecCommand", \
              }, \
            { "button12"                  : ["clicked", self.BrowseForJobTemplate], \
              "button70"                  : ["clicked", self.RemoveMachine], \
              "button69"                  : ["clicked", self.CloneMachine], \
              "button67"                  : ["clicked", self.NewMachine], \
#             "treeview2"                 : ["button-press-event", self.ResListTreeViewButtonPress], \
#             "treeview2"                 : ["popup-menu", self.SelectResource], \
              "treeview2"                 : ["changed", self.SelectResource, "get_selection"], \
              "button71"                  : ["clicked", self.AddQueueToMachine], \
              "button72"                  : ["clicked", self.RemoveQueueFromMachine], \
              "button73"                  : ["clicked", self.AddAccountToMachine], \
              "button74"                  : ["clicked", self.RemoveAccountFromMachine], \
              "combobox1"                 : ["changed", self.FillMachAccDialogData], \
              }, \
            "treeview2" )
        self.PopulateResListTreeView(self.MachineDialog, self.Data.Machines)
        self.AuthenticationDialog = self.CreateDialog("AuthenticationDialog", \
            { "label32"                   : "LoginHost", \
              "entry14"                   : "RemoteUserid", \
              "label34"                   : "AuthDataName", \
              "entry15"                   : "AuthData", \
              "checkbutton1"              : "StoreAuthData", \
              } )
        self.HostCertAuthenticationDialog = self.CreateDialog("HostCertAuthenticationDialog", \
            { "label291"                  : "LoginHost", \
              "entry95"                   : "RemoteUserid", \
              "label292"                  : "AuthDataName", \
              "entry96"                   : "AuthData", \
              } )
        self.CertAuthenticationDialog = self.CreateDialog("CertAuthenticationDialog", \
            { "label288"                  : "Realm", \
              "entry93"                   : "RemoteUserid", \
              "label289"                  : "AuthDataName", \
              "entry94"                   : "AuthData", \
              } )
        self.GridAuthenticationDialog = self.CreateDialog("GridAuthenticationDialog", \
            { "label317"                  : "AuthDataName", \
              "entry106"                  : "AuthData", \
              } )
        self.OutputWindow = self.CreateDialog("OutputWindow", \
            { "textview11"                : "String", \
              } )
        self.FileViewWindow = self.CreateDialog("FileViewWindow", \
            { "textview13"                : "String", \
              } )
        self.SubmitDialog = self.CreateDialog("SubmitDialog", \
            { "textview12"                : "String", \
              } )
        self.wTree.splash.hide()
        self.UpdateTimeout         = -1
        self.UpdateLock            = False
        self.TotalFilesSent        = 0
        self.TotalFilesToSend      = 1
        self.FileTransferCancelled = False
        self.ProgressWindow              = gtk.glade.XML(GladeFile, "FileTransferProgressWindow")
        self.ProgressWindow.window       = self.ProgressWindow.get_widget("FileTransferProgressWindow")
        self.ProgressWindow.window.hide()
        self.ProgressWindow.progressbar  = self.ProgressWindow.get_widget("progressbar1")
        self.ProgressWindow.name         = self.ProgressWindow.get_widget("label294")
        self.ProgressWindow.label        = self.ProgressWindow.get_widget("label304")
        self.ProgressWindow.cancelbutton = self.ProgressWindow.get_widget("button91")
        self.RemoteCommandCancelled = False
        self.CmdProgressWindow              = gtk.glade.XML(GladeFile, "RemoteCommandProgressWindow")
        self.CmdProgressWindow.window       = self.CmdProgressWindow.get_widget("RemoteCommandProgressWindow")
        self.CmdProgressWindow.window.hide()
        self.CmdProgressWindow.progressbar  = self.CmdProgressWindow.get_widget("progressbar2")
        self.CmdProgressWindow.name         = self.CmdProgressWindow.get_widget("label361")
        self.CmdProgressWindow.label        = self.CmdProgressWindow.get_widget("label362")
        self.CmdProgressWindow.cancelbutton = self.CmdProgressWindow.get_widget("button100")

    def SplashScreen (self):
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        window.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_SPLASHSCREEN)
        window.set_decorated(False)
        window.set_gravity(gtk.gdk.GRAVITY_CENTER)
        window.set_position(gtk.WIN_POS_CENTER)
        hbox = gtk.HBox()
        window.add(hbox)
        image = gtk.Image()
        image.set_from_file(os.path.join(PixmapDirectory, "squid_circuit_2.png"))
        hbox.pack_start(image)
        image.show()
        hbox.show()
        window.show()
        return window

    def HelpWindow (self, widget, data=None):
        webbrowser.open("file://"+DocDirectory+"/index.html")
    
    def RunAuthenticationDialog (self, machine, method):
        if (method.Name == "local"):
            return SM_OK
        if ((machine.RemoteUserid == "") or (machine.LoginHost == "")) and (not machine.AlreadyPrompted):
            result = self.QuestionDialog('Machine profile "'+machine.Name+'" is missing a remote userid or login host '+\
                                         'specification.  This usually causes trouble.  Are you sure you want to proceed?')
            if (result == gtk.RESPONSE_CANCEL):
                return SM_CANCEL
        result = SM_OK
        # ticket/proxy check (so that we re-prompt when ticket/proxy has expired)
        if (method.Type == SM_AUTH_KERBEROS):
            pass  # to do... klist doesn't seem to let you check just one ticket?
        elif (method.Type == SM_AUTH_GSI):
            output = ""
            try:
                process = pexpect.spawn("grid-proxy-info -exists -identity")
                process.expect(pexpect.EOF)
                current_proxy_info = process.before
                process.close(wait=1)
                if process.exitstatus or (current_proxy_info.count(machine.GridCertSubject) == 0):
                    machine.AlreadyPrompted = False
                else:
                    machine.AlreadyPrompted = True
            except pexpect.ExceptionPexpect, e:
                result = SM_FAIL
                output = output + "\n%s" % e
                machine.AlreadyPrompted = False
                self.NotifyDialog("Error while issuing grid-proxy-info command.  Is the Globus" +\
                                  " toolkit installed and in your default path?")
                return result
        if ((method.Type == SM_AUTH_KERBEROS)     and (not machine.AlreadyPrompted)) or \
           ((method.Type == SM_AUTH_GSI)          and (not machine.AlreadyPrompted)) or \
           ((method.Type == SM_AUTH_KEYAGENT)     and (not machine.AlreadyPrompted)) or \
           ((method.Type == SM_AUTH_PASSWORD)     and (not (machine.StoreAuthData and \
                                                            machine.AlreadyPrompted))):
            if (method.Type == SM_AUTH_KERBEROS):
                dialog = self.CertAuthenticationDialog
            elif (method.Type == SM_AUTH_KEYAGENT):
                dialog = self.HostCertAuthenticationDialog
            elif (method.Type == SM_AUTH_GSI):
                if self.Data.CurrentUser.UseMyProxy:
                    # need to move this code into a separate proxy auth method
                    dTree = gtk.glade.XML(GladeFile, "MyProxyAuthenticationDialog")
                    dTree.dialog = dTree.get_widget("MyProxyAuthenticationDialog")
                    dTree.get_widget("label364").set_text(self.Data.CurrentUser.MyProxyServer)
                    dTree.get_widget("label368").set_text(self.Data.CurrentUser.MyProxyAccount)
                    dTree.get_widget("label370").set_text(machine.GridCertSubject)
                    dTree.dialog.present()
                    result = dTree.dialog.run()
                    myproxy_passphrase = dTree.get_widget("entry113").get_text()
                    dTree.get_widget("entry113").set_text("")
                    dTree.dialog.destroy()
                    if (result == gtk.RESPONSE_CANCEL):
                        return SM_CANCEL
                    command = "myproxy-logon -s "+self.Data.CurrentUser.MyProxyServer+\
                              " -l "+self.Data.CurrentUser.MyProxyAccount+\
                              " -t "+str(self.Data.CurrentUser.GridProxyLifetime)+\
                              " -k \""+machine.GridCertSubject+"\""
                    output = ""
                    try:
                        process = pexpect.spawn(command)
                        index = process.expect(["Enter MyProxy Pass Phrase:", pexpect.EOF])
                        output = process.before + process.after
                        if (index == 0):
                            process.sendline(myproxy_passphrase)
                            process.expect(pexpect.EOF)
                            output = output + process.before + process.after
                        myproxy_passphrase = ""
                        process.close(wait=1)
                        if process.exitstatus:
                            result = SM_FAIL
                        else:
                            result = SM_CONTINUE
                    except pexpect.ExceptionPexpect, e:
                        result = SM_FAIL
                        output = output + "\n%s" % e
                        self.NotifyDialog("Error while issuing myproxy-logon command.  Are MyProxy" +\
                                          " client tools installed and in your default path?")
                        return result
                    # end of proxy auth code
                else:
                    dialog = self.GridAuthenticationDialog
                    certfile, keyfile = self.Data.LocateGridIdentityFiles(machine.GridCertSubject, \
                                                                          self.Data.CurrentUser.GridCertDir)
                    if (not certfile) or (not keyfile):
                        self.NotifyDialog("Could not obtain X.509 certificate information for identity\n" + \
                                          machine.GridCertSubject+"\n"+ \
                                          "Please check your grid and certificate settings and" + \
                                          " try again.")
                        return SM_FAIL
                    dialog.get_widget("label318").set_text("%s" % machine.GridCertSubject)
                    machine.GridCertFile = certfile
                    machine.GridKeyFile  = keyfile
            else:
                dialog = self.AuthenticationDialog
            if (result != SM_FAIL) and (result != SM_CONTINUE):
                self.SetDialogData(dialog, machine)
                dialog.dialog.present()
                result = dialog.dialog.run()
                dialog.dialog.hide()
                if (result != gtk.RESPONSE_CANCEL):
                    self.GetDialogData(dialog, machine, clear=True)
                    if ((method.Type == SM_AUTH_KERBEROS) or (method.Type == SM_AUTH_GSI) or \
                        (method.Type == SM_AUTH_KEYAGENT)):
                        # run init command...
                        test_result, output = method.GetCredentials(machine.RemoteUserid, \
                                                                    machine.LoginHost, machine.Realm, \
                                                                    machine.AuthData, \
                                                                    machine, self.Data)
                    else:
                        # test connection in some other way?
                        command = OSTypes[machine.OSType]["TestCommand"]
                        test_result, output = self.Data.RemoteExec(machine.Name, "", command, store_auth=True)
                    if (test_result == SM_OK):
                        result = SM_OK
                        machine.AlreadyPrompted = True
                    else:
                        result = SM_FAIL
                        machine.AlreadyPrompted = False
                        machine.StoreAuthData = False
                else:
                    result = SM_CANCEL
        while gtk.events_pending():
            gtk.main_iteration()
        if (result == SM_FAIL):
            if self.Data.CurrentUser.VerboseLogging:
                self.ViewWindow("Remote authentication failed", "Remote authentication failed; external command"+ \
                                " output appears below:\n\n"+output)
            else:
                self.NotifyDialog("Remote authentication failed.  Either the authentication data supplied"+\
                              	  " were incorrect, or the remote machine did not respond as expected.")
        return result

    def AddTreeViewColumn (self, name, text_column, foreground_column, foreground_set_column, \
                           font_column, min_width, max_width, expand):
        self.tvcolumn  = gtk.TreeViewColumn(name)
        self.treeview.append_column(self.tvcolumn)
        self.cell      = gtk.CellRendererText()
        self.tvcolumn.pack_start(self.cell, True)
        self.tvcolumn.add_attribute(self.cell, "text", text_column)
        self.tvcolumn.add_attribute(self.cell, "foreground-gdk", foreground_column)
        self.tvcolumn.add_attribute(self.cell, "foreground-set", foreground_set_column)
        self.tvcolumn.add_attribute(self.cell, "font-desc", font_column)
        self.tvcolumn.set_min_width(min_width)
        self.tvcolumn.set_max_width(max_width)
        self.tvcolumn.set_expand(expand)
#       self.tvcolumn.set_resizable(True)
    
    def ToggleShowDescription (self, widget, data=None):
        self.Data.CurrentUser.TreeViewShowDescription = not self.Data.CurrentUser.TreeViewShowDescription
        self.SetTreeViewColumns()
        return True
        
    def ToggleShowStatus (self, widget, data=None):
        self.Data.CurrentUser.TreeViewShowStatus = not self.Data.CurrentUser.TreeViewShowStatus
        self.SetTreeViewColumns()
        return True
        
    def ToggleShowModifiedDate(self, widget, data=None):
        self.Data.CurrentUser.TreeViewShowModifiedDate = not self.Data.CurrentUser.TreeViewShowModifiedDate
        self.SetTreeViewColumns()
        return True
        
    def SetTreeViewColumns (self):
        for tvcolumn in self.treeview.get_columns():
            self.treeview.remove_column(tvcolumn)
        self.AddTreeViewColumn("Name", 0, 5, 6, 7, 50, 250, True)
        if self.Data.CurrentUser.TreeViewShowDescription:
            self.AddTreeViewColumn("Description", 2, 5, 6, 7, 100, 300, True)
        if self.Data.CurrentUser.TreeViewShowStatus:
            self.AddTreeViewColumn("Status", 3, 5, 6, 7, 150, 400, True)
        if self.Data.CurrentUser.TreeViewShowModifiedDate:
            self.AddTreeViewColumn("Date last modified", 4, 5, 6, 7, 50, 175, True)

    def MakeTreeViewConsistent (self):
        widget = self.wTree.get_widget("show_description1")
        self.Data.CurrentUser.TreeViewShowDescription = widget.get_active()
        widget = self.wTree.get_widget("show_status1")
        self.Data.CurrentUser.TreeViewShowStatus = widget.get_active()
        widget = self.wTree.get_widget("show_modification_date1")
        self.Data.CurrentUser.TreeViewShowModifiedDate = widget.get_active()

    def ExpandTreeView (self, widget, data=None):
        self.treeview.expand_all()
        return True
        
    def CollapseTreeView (self, widget, data=None):
        self.treeview.collapse_all()
        return True

    def FillMachAccDialogData (self, widget, data=None):
        self.GetDialogData(self.MachineDialog, self.Data.Machines.Current)
        method = self.Data.AccessMethodByName(self.Data.Machines.Current.AccessMethod)
        self.Data.Machines.Current.AuthInitCommand = \
                method.AuthInitCommand.String
        self.Data.Machines.Current.RemoteExecCommand = \
                method.RemoteExecCommand.String
        self.Data.Machines.Current.FileUploadCommand = \
                method.FileUploadCommand.String
        self.Data.Machines.Current.FileDownloadCommand = \
                method.FileDownloadCommand.String
        self.Data.Machines.Current.AuthDataName = \
                method.AuthDataName
        self.Data.Machines.Current.AuthDataPrompt = \
                method.AuthDataPrompt
        self.Data.Machines.Current.AccessMethodType = \
                method.Type
        self.Data.Machines.Current.StoreAuthData = \
                method.StoreAuthData
        self.Data.Machines.Current.ThirdPartyTransferCmd = \
                method.ThirdPartyTransferCmd
        self.SetMachineDialogData()
    
    def FillExpExeMachDialogData (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        Mach = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        App  = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        NewChoices = None
        if Mach:
            self.Data.CurrentExperiment.ExecQueue      = Mach.DefaultQueue
            self.Data.CurrentExperiment.ExecAccount    = Mach.DefaultAccount
            self.Data.CurrentExperiment.RemoteBuildDir = Mach.RemoteBuildRootDir
            self.Data.CurrentExperiment.Tiling         = Mach.DefaultCPUsPerNode
            self.Data.CurrentExperiment.MemoryPerNode  = Mach.MemoryPerNode
            CmdSep  = OSTypes[Mach.OSType]["CommandSeparator"]
            if App:
                self.Data.CurrentExperiment.ConfigCommand = App.ConfigCommand
                if App.BuildExecutable:
                    self.Data.CurrentExperiment.RemoteBuildDir = \
                        self.Data.RemotePath([self.Data.CurrentExperiment.RemoteBuildDir, \
                                              self.Data.MungeStringToPath(App.Name)], \
                                             Mach.Name)
                    self.Data.CurrentExperiment.BuildCommand = App.BuildCommand
                    if App.MoveExecutable:
                        self.Data.CurrentExperiment.BuildCommand = \
                            self.Data.CurrentExperiment.BuildCommand + CmdSep + \
                            Mach.GetFileMoveCommand(App.ExecutableName, Mach.RemoteExecDir)
                        self.Data.CurrentExperiment.Executable = \
                            self.Data.RemotePath([Mach.RemoteExecDir, App.ExecutableName], \
                                                 Mach.Name)
                    else:
                        self.Data.CurrentExperiment.Executable = \
                            self.Data.RemotePath([self.Data.CurrentExperiment.RemoteBuildDir, \
                                                  App.ExecutableName], Mach.Name)
                else:
                    self.Data.CurrentExperiment.RemoteBuildDir = ""
                    self.Data.CurrentExperiment.BuildCommand   = ""
                    self.Data.CurrentExperiment.Executable = App.ExecutableName
            NewChoices = { "ExecQueue" : Mach.Queues, "ExecAccount" : Mach.Accounts }
        elif App:
            self.Data.CurrentExperiment.ConfigCommand = App.ConfigCommand
            if App.BuildExecutable:
                self.Data.CurrentExperiment.BuildCommand = App.BuildCommand
            else:
                self.Data.CurrentExperiment.BuildCommand = ""
            self.Data.CurrentExperiment.Executable    = App.ExecutableName
            NewChoices = {}
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment, \
                           NewChoices)
    
    def AddQueueToMachine (self, widget, data=None):
        combobox  = self.MachineDialog.get_widget("comboboxentry24")
        entry     = combobox.child
        queue     = entry.get_text()
        treemodel = combobox.get_model()
        treeiter  = treemodel.get_iter_first()
        found = False
        while (treeiter) and (not found):
            found = (treemodel.get_value(treeiter, 0) == queue)
            treeiter = treemodel.iter_next(treeiter)
        if (not found):
            combobox.append_text(queue)
            self.Data.Machines.Current.Queues.append(queue)
        
    def AddAccountToMachine (self, widget, data=None):
        combobox  = self.MachineDialog.get_widget("comboboxentry4")
        entry     = combobox.child
        account   = entry.get_text()
        treemodel = combobox.get_model()
        treeiter  = treemodel.get_iter_first()
        found = False
        while (treeiter) and (not found):
            found = (treemodel.get_value(treeiter, 0) == account)
            treeiter = treemodel.iter_next(treeiter)
        if (not found):
            combobox.append_text(account)
            self.Data.Machines.Current.Accounts.append(account)
        
    def RemoveQueueFromMachine (self, widget, data=None):
        combobox  = self.MachineDialog.get_widget("comboboxentry24")
        entry     = combobox.child
        queue     = entry.get_text()
        treemodel = combobox.get_model()
        treeiter  = treemodel.get_iter_first()
        found = False
        i = 0
        while (treeiter) and (not found):
            found = (treemodel.get_value(treeiter, 0) == queue)
            treeiter = treemodel.iter_next(treeiter)
            i = i + 1
        if found:
            entry.set_text("")
            combobox.remove_text(i-1)
            self.Data.Machines.Current.Queues.remove(queue)
        
    def RemoveAccountFromMachine (self, widget, data=None):
        combobox  = self.MachineDialog.get_widget("comboboxentry4")
        entry     = combobox.child
        account   = entry.get_text()
        treemodel = combobox.get_model()
        treeiter  = treemodel.get_iter_first()
        found = False
        i = 0
        while (treeiter) and (not found):
            found = (treemodel.get_value(treeiter, 0) == account)
            treeiter = treemodel.iter_next(treeiter)
            i = i + 1
        if found:
            entry.set_text("")
            combobox.remove_text(i-1)
            self.Data.Machines.Current.Accounts.remove(account)
        
    def FillExpAppDialogData (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        App = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        if App:
            self.Data.CurrentExperiment.ConfigCommand = App.ConfigCommand
            if App.BuildExecutable:
                self.Data.CurrentExperiment.BuildCommand = App.BuildCommand
            else:
                self.Data.CurrentExperiment.BuildCommand = ""
            self.Data.CurrentExperiment.Executable    = App.ExecutableName
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)

    def BrowseForAppSourceDir (self, widget, data=None):
        self.GetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
        result, filename = self.FileChooserDialog(self.Data.Applications.Current.LocalSourcePath, \
                                                  "FolderChooserDialog")
        if (result == gtk.RESPONSE_OK):
            self.Data.Applications.Current.LocalSourcePath = filename
        self.SetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
    
    def RemoveApplication (self, widget, data=None):
        dTree  = gtk.glade.XML(GladeFile, "DeleteConfirmDialog")
        dialog = dTree.get_widget("DeleteConfirmDialog")
        label  = dTree.get_widget("label189")
        label.set_text('Application "' + self.Data.Applications.Current.Name + '"')
        result = dialog.run()
        dialog.destroy()
        if (result == gtk.RESPONSE_OK):
            self.Data.Applications.Delete(self.Data)
            self.SetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
            self.PopulateResListTreeView(self.ApplicationDialog, self.Data.Applications)
            self.Data.Changed = True

    def NewApplication (self, widget, data=None):
        self.Data.Applications.New(self.Data)
        self.SetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
        self.PopulateResListTreeView(self.ApplicationDialog, self.Data.Applications)
        self.Data.Changed = True

    def CloneApplication (self, widget, data=None):
        self.Data.Applications.Clone()
        self.SetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
        self.PopulateResListTreeView(self.ApplicationDialog, self.Data.Applications)
        self.Data.Changed = True

    def BrowseForJobTemplate (self, widget, data=None):
        self.GetDialogData(self.MachineDialog, self.Data.Machines.Current)
        result, filename = self.FileChooserDialog(self.Data.Machines.Current.JobTemplateFile, \
                                                  "FileChooserDialog")
        if (result == gtk.RESPONSE_OK):
            self.Data.Machines.Current.JobTemplateFile = filename
        self.SetDialogData(self.MachineDialog, self.Data.Machines.Current)
    
    def RemoveMachine (self, widget, data=None):
        dTree  = gtk.glade.XML(GladeFile, "DeleteConfirmDialog")
        dialog = dTree.get_widget("DeleteConfirmDialog")
        label  = dTree.get_widget("label189")
        label.set_text('Machine "' + self.Data.Machines.Current.Name + '"')
        result = dialog.run()
        dialog.destroy()
        if (result == gtk.RESPONSE_OK):
            self.Data.Machines.Delete(self.Data)
            self.SetMachineDialogData()
            self.PopulateResListTreeView(self.MachineDialog, self.Data.Machines)
            self.Data.Changed = True
    
    def NewMachine (self, widget, data=None):
        self.Data.Machines.New(self.Data)
        self.SetMachineDialogData()
        self.PopulateResListTreeView(self.MachineDialog, self.Data.Machines)
        self.Data.Changed = True

    def CloneMachine (self, widget, data=None):
        self.Data.Machines.Clone()
        self.SetMachineDialogData()
        self.PopulateResListTreeView(self.MachineDialog, self.Data.Machines)
        self.Data.Changed = True
       
    def ResListTreeViewButtonPress (self, widget, event):
        if (event.button == 1):
            treeselection = widget.get_selection()
            treepath = widget.get_path_at_pos(int(event.x), int(event.y))
            if treepath:
                treepath = treepath[0]
                treeselection.unselect_all()
                treeselection.select_path(treepath)
                self.SelectResource(widget, event)
            return True
        else:
            return False
        
    def SelectResource (self, widget, data=None):
#       treeselection = widget.get_selection()
        treeselection = widget
        (treemodel, treeiter) = treeselection.get_selected()
        if (treemodel == self.ApplicationDialog.treestore):
            dialog   = self.ApplicationDialog
            reslist  = self.Data.Applications
        else:   # machine list
            dialog   = self.MachineDialog
            reslist  = self.Data.Machines
        resource = None
        for entry in reslist.Entries:
            if (dialog.treestore.get_value(treeiter, 1) == entry):
                resource = entry
        if resource:
            reslist.Current = resource
            if (treemodel == self.MachineDialog.treestore):
                self.SetMachineDialogData()
            else:
                self.SetDialogData(dialog, reslist.Current)
            return True
        else:
            return False
        
    def PopulateResListTreeView (self, dialog, reslist):
        if dialog.treestore:
            for widget_name in dialog.handler_ids.keys():
                widget = dialog.handler_ids[widget_name][0]
                widget.handler_block(dialog.handler_ids[widget_name][1])
            dialog.treestore.clear()
            curiter = None
            for Entry in reslist.Entries:
                treeiter = dialog.treestore.append([Entry.Name, Entry])
                if (Entry == reslist.Current):
                    curiter = treeiter
            treeselection = dialog.treeview.get_selection()
            treeselection.unselect_all()
            if curiter:
                treeselection.select_iter(curiter)
            for widget_name in dialog.handler_ids.keys():
                widget = dialog.handler_ids[widget_name][0]
                widget.handler_unblock(dialog.handler_ids[widget_name][1])
        
    def GenerateRuns (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        machine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        if not machine:
            self.NotifyDialog("Please choose an execution machine before generating runs.")
            return True
        app = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        if not app:
            self.NotifyDialog("Please choose an application before generating runs.")
            return True
        result = self.Data.CurrentExperiment.CheckData(self.Data)
        if (result == SM_OK):
            result = gtk.RESPONSE_OK
            if not self.Data.CurrentExperiment.ParmFileTemplate:
                result = self.QuestionDialog("Proceed without a parameter file template?")
            if (result == gtk.RESPONSE_OK):
                NewChildren = self.Data.CurrentExperiment.AddChildren(self.Data)
                if (None in NewChildren):
                    self.NotifyDialog("Could not generate runs.  Is the parameter file template missing?")
                else:
                    for Run in NewChildren:
                        self.AddTreeStoreRow(self.Data.CurrentExperiment, Run.Name, Run)
        else:
            self.NotifyDialog("Error in parameter range settings.")

    def CancelFileTransfer (self, widget, data=None):
        self.FileTransferCancelled = True
        self.Data.CancelRemoteExec()
        return True

    def CancelRemoteCommand (self, widget, data=None):
        self.RemoteCommandCancelled = True
        self.Data.CancelRemoteExec()
        return True

    def OpenRemoteCommandProgressWindow (self, title, command, parent_dialog):
        self.LockUpdates()
        self.TendrilInProgress = True
        self.RemoteCommandCancelled = False
        self.CmdProgressWindow.window.set_title(title)
        self.CmdProgressWindow.cancelbutton.connect("clicked", self.CancelRemoteCommand)
        self.CmdProgressWindow.progressbar.set_fraction(0.0)
        self.CmdProgressWindow.label.set_text(command)
        self.CmdProgressWindow.progressbar.set_text(" ")
        self.CmdProgressWindow.window.show()
        self.CmdProgressWindow.window.present()
        gobject.timeout_add(1000, self.PulseProgressBar, self.CmdProgressWindow.progressbar)
        while gtk.events_pending():
            gtk.main_iteration()

    def CloseRemoteCommandProgressWindow (self):
        self.TendrilInProgress = False
        self.CmdProgressWindow.window.hide()
        self.UnlockUpdates()

    def PulseProgressBar (self, progressbar):
        if self.TendrilInProgress:
            progressbar.pulse()
            return True
        else:
            return False

    def OpenFileTransferProgressWindow (self, title, parent_dialog):
        self.LockUpdates()
        self.ProgressWindow.window.set_title(title)
        self.ProgressWindow.cancelbutton.connect("clicked", self.CancelFileTransfer)
        self.ProgressWindow.progressbar.set_fraction(0.0)
        self.ProgressWindow.label.set_text(" ")
        self.ProgressWindow.progressbar.set_text(" ")
        self.ProgressWindow.window.show()
        self.ProgressWindow.window.present()

    def UpdateFileTransferProgressWindow (self, file, fraction, text=None):
        path, filename = os.path.split(file)
        self.ProgressWindow.label.set_text(filename)
        if (fraction >= 0.0) and (fraction <= 1.0):
            self.ProgressWindow.progressbar.set_fraction(fraction)
        if text:
            self.ProgressWindow.progressbar.set_text(text)
        while gtk.events_pending():
            gtk.main_iteration()

    def CloseFileTransferProgressWindow (self):
        self.ProgressWindow.window.hide()
        self.UnlockUpdates()

    def TransferFiles (self, SourceHost, SourceFileList, DestHost, DestPath, NewProgressWindow=True, \
                       ParentDialog=None, DestIsDir=False):
        if (SourceHost != SM_LOCALHOST):
            if not (SourceHost in self.Data.Machines.Entries):
                self.NotifyDialog('Unknown data source host.  File transfer aborted.')
                return SM_FAIL
            SourceMethod = self.Data.AccessMethodByName(SourceHost.AccessMethod)
            if not SourceMethod:
                self.NotifyDialog('No access method configured for host "%s".  File transfer aborted.' % SourceHost.Name)
                return SM_FAIL
        else:
            SourceMethod = self.Data.LocalMethod
        if (DestHost != SM_LOCALHOST):
            if not (DestHost in self.Data.Machines.Entries):
                self.NotifyDialog('Unknown data destination host.  File transfer aborted.')
                return SM_FAIL
            DestMethod = self.Data.AccessMethodByName(DestHost.AccessMethod)
            if not DestMethod:
                self.NotifyDialog('No access method configured for host "%s".  File transfer aborted.' % DestHost.Name)
                return SM_FAIL
        else:
            DestMethod = self.Data.LocalMethod
        BackgroundTransfer = False
        ThirdPartyTransfer = False
        if (SourceHost != DestHost) and (SourceMethod.Name != "local") and (DestMethod.Name != "local"):
            if (SourceMethod != DestMethod) or (not SourceMethod.ThirdPartyTransferCmd):
                self.NotifyDialog('Third-party and background file transfers are only supported '+ \
                                  'if both source and destination hosts are configured for Grid-protocol '+ \
                                  'access.')
                return SM_FAIL
            else:
                ThirdPartyTransfer = True
        if NewProgressWindow and (not BackgroundTransfer):
            self.FileTransferCancelled = False
            self.OpenFileTransferProgressWindow("File transfer", ParentDialog)
            self.ProgressWindow.name.set_text("Transferring files...")
            self.Data.LogMessage("Transferring files...")
            self.TotalFilesSent = 0
            self.TotalFilesToSend = len(SourceFileList)
        try:
            for SourceFile in SourceFileList:
                SourceAuthResult = self.RunAuthenticationDialog(SourceHost, SourceMethod)
                if (SourceAuthResult == SM_FAIL):   raise RemoteExecError
                if (SourceAuthResult == SM_CANCEL): raise UserCancelEvent
                DestAuthResult   = self.RunAuthenticationDialog(DestHost, DestMethod)
                if (DestAuthResult == SM_FAIL):     raise RemoteExecError
                if (DestAuthResult == SM_CANCEL):   raise UserCancelEvent
                if not BackgroundTransfer:
                    self.UpdateFileTransferProgressWindow(SourceFile, float(self.TotalFilesSent)/float(self.TotalFilesToSend), \
                                                          "%d / %d" % (self.TotalFilesSent+1, self.TotalFilesToSend))
                if (len(SourceFileList) > 1) or DestIsDir:       # if multiple source files, interpret destination as a directory
                    if (SourceHost == SM_LOCALHOST):
                        PathList = self.Data.PathList(SourceFile)
                    else:
                        PathList = self.Data.RemotePathList(SourceFile, SourceHost.Name)
                    if (DestHost == SM_LOCALHOST):
                        DestFile = os.path.join(DestPath, PathList[-1])
                    else:
                        DestFile = self.Data.RemotePath([DestPath, PathList[-1]], DestHost.Name)
                else:
                    DestFile = DestPath
                result, output = self.Data.TransferFile(SourceHost, SourceFile, DestHost, DestFile)
                while gtk.events_pending():
                    gtk.main_iteration()
                if (result == SM_FAIL):             raise RemoteExecError
                if self.FileTransferCancelled:      raise UserCancelEvent
                self.TotalFilesSent += 1
        except UserCancelEvent:
            self.Data.LogMessage("File transfer cancelled by user.")
            if NewProgressWindow and (not BackgroundTransfer):
                self.CloseFileTransferProgressWindow()
            return SM_CANCEL
        except RemoteExecError:
            self.Data.LogMessage("Error while transferring files.")
            if self.Data.CurrentUser.ShowOutputOnError:
                self.ViewWindow("External command error", \
                                "File transfer failed.  Command output appears below.\n\n" + output)
#            else:
#                self.NotifyDialog("External command error while transferring files.")
            if NewProgressWindow and (not BackgroundTransfer):
                self.CloseFileTransferProgressWindow()
            return SM_FAIL
        if NewProgressWindow and (not BackgroundTransfer):
            self.CloseFileTransferProgressWindow()
        return result

    def MakeRemoteDirectory (self, machine, path):
        if (machine != SM_LOCALHOST):
            if not (machine in self.Data.Machines.Entries):
                self.NotifyDialog("Could not make remote directory; machine not found in list.")
                return (SM_FAIL, "")
        method = self.Data.AccessMethodByName(machine.AccessMethod)
        if not method:
            self.NotifyDialog("Could not make remote directory; machine %s is not configured with"+\
                              " an access method.")
            return (SM_FAIL, "")
        result = self.RunAuthenticationDialog(machine, method)
        output = ""
        if (result == SM_OK):
            result, output = self.Data.MakeRemoteDirectory(machine, path)
            if (result != SM_OK):
                if self.Data.CurrentUser.ShowOutputOnError:
                    self.ViewWindow("External command error", \
                                    "Remote directory creation failed.  "+\
                                    "Command output appears below.\n\n" + output)
#                else:
#                    self.NotifyDialog("Could not make remote directory.")
        return result, output
    
    def RemoteExec (self, machine, startdir, command):
        if not command:
            self.NotifyDialog("Invalid remote command.  Please check settings and try again.")  
            return (SM_FAIL, "RemoteExec:  invalid command")
        if not machine:
            self.NotifyDialog("Invalid remote machine.  Please check settings and try again.")
            return (SM_FAIL, "RemoteExec:  invalid machine")
        method  = self.Data.AccessMethodByName(machine.AccessMethod)
        if not method:
            self.NotifyDialog("Remote machine has not been configured with an access method." + \
                              "  Please check settings and try again.")
            return (SM_FAIL, "RemoteExec:  invalid access method")
        result = self.RunAuthenticationDialog(machine, method)
        output = ""
        if (result == SM_OK):
            result, output = self.Data.RemoteExec(machine.Name, startdir, command)
            if (result != SM_OK):
                if self.Data.CurrentUser.ShowOutputOnError:
                    self.ViewWindow("External command error", \
                                    "Remote execution failed.  "+\
                                    "Command output appears below.\n\n" + output)
#                else:
#                    self.NotifyDialog("Remote command execution failed.")
        return result, output

    def UploadSourceFiles (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        machine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        if (not machine):
            self.NotifyDialog("Please choose an execution machine before attempting to upload source files.")
            return True
        app     = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        if (not app):
            self.NotifyDialog("Please choose an application before attempting to upload source files.")
            return True
        #
        # Create the remote build directory
        #
        src     = app.LocalSourcePath
        dest    = self.Data.CurrentExperiment.RemoteBuildDir
        result, output = self.MakeRemoteDirectory(machine, dest)
        #
        # Transfer tendril executable if necessary
        #
        self.FileTransferCancelled = False
        if ((result == SM_OK) and (self.Data.CurrentUser.UseTendril)):
            result, RemoteTendrilPath = self.ExtendTendril(machine)
        #
        # Upload source code.  If using tendril, first determine which files and directories have changed
        # on the local machine.
        #
        if (result == SM_OK):
            self.Data.LogMessage("Source code upload:")
            self.OpenFileTransferProgressWindow("Source upload", self.wTree.window)
            try:
                save_cwd = os.getcwd()
                os.chdir(src)
                #
                # Get list of files and directories to transfer
                #
                if self.Data.CurrentUser.UseTendril:
                    self.ProgressWindow.name.set_text("Comparing file lists with tendril...")
                    self.Data.LogMessage("Comparing file lists with tendril...")
                    self.TendrilInProgress = True
                    gobject.timeout_add(100, self.PulseProgressBar, self.ProgressWindow.progressbar)
                    while gtk.events_pending():
                        gtk.main_iteration()
                    try:
                        process = pexpect.spawn(TendrilPath + " md5walk .")
                        self.Data.LiveRemoteProcess = process
                        process.expect(pexpect.EOF, timeout=SM_TIMEOUT)
                        local_output = "%s%s" % (process.before, process.after)
                        process.close(wait=1)
                        self.Data.LiveRemoteProcess = None
                    except pexpect.ExceptionPexpect, e:
                        self.NotifyDialog("Error while invoking tendril on your local machine.")
                        raise RemoteExecError
                    while gtk.events_pending():
                        gtk.main_iteration()
                    if self.FileTransferCancelled:
                        raise UserCancelEvent
                    result, remote_output = self.RemoteExec(machine, dest, RemoteTendrilPath+" md5walk .")
                    while gtk.events_pending():
                        gtk.main_iteration()
                    if self.FileTransferCancelled:
                        raise UserCancelEvent
                    # Here need to compare the two output lists to determine pruned file_tree to send
                    # Send each file in local_output but not in remote_output
                    # Send each file listed in both for which MD5's are not the same
                    # Do not send any others
                    # If any errors, revert to sending all files
                    if (result == SM_OK):
                        file_tree = self.Data.MD5Diff(local_output, remote_output)
                    else:
                        file_tree = os.walk('.')
                    self.TendrilInProgress = False
                else:
                    file_tree = os.walk('.')
                #
                # Prune the file/directory list of duplicates and CVS directories.  Create separate lists of the
                # files to send and directories to make.
                #
                dirs  = [ ]
                paths = [ ]
                total_size = 0
                total_dirs = 0
                for root, dirnames, filenames in file_tree:
                    RootPathList = [self.Data.CurrentExperiment.RemoteBuildDir]
                    RootPathList.extend(self.Data.PathList(root))
                    if ("CVS" in dirnames):
                        dirnames.remove("CVS")
                    if not ("CVS" in RootPathList):
                        if (dirnames == [ ]):
                            path = self.Data.RemotePath(RootPathList, machine.Name)
                            dirs.append(path)
                            total_dirs += 1
                        for file in filenames:
                            path = os.path.join(root, file)
                            if (path[-1] == "\r"):
                                path = path[:-1]
                            size = os.path.getsize(path)
                            paths.append(path)
#                            paths.append((path, size))
                            total_size += size
                    if self.FileTransferCancelled:
                        raise UserCancelEvent
                self.Data.LogMessage("Total paths to make:     %d" % total_dirs)
                self.Data.LogMessage("Total bytes to transfer: %d" % total_size)
                #
                # Create the necessary directories on the remote machine.
                #
                result = SM_OK
                self.ProgressWindow.name.set_text("Replicating directory structure...")
                self.Data.LogMessage("Replicating directory structure...")
                if self.Data.CurrentUser.UseTendril and (len(dirs) > 0):
                    self.TendrilInProgress = True
                    gobject.timeout_add(100, self.PulseProgressBar, self.ProgressWindow.progressbar)
                    while gtk.events_pending():
                        gtk.main_iteration()
                    f = open(TempFile, 'w')
                    for dir in dirs:
                        f.write(dir + "\n")
                    f.close()
                    result = self.TransferFiles(SM_LOCALHOST, [TempFile], machine, \
                                                self.Data.RemotePath([dest, "teuthis.dat"], machine.Name), False, None)
                    if (result == SM_OK):
                        result, output = self.RemoteExec(machine, dest, RemoteTendrilPath+" mmkdir teuthis.dat")
                    self.TendrilInProgress = False
                    while gtk.events_pending():
                        gtk.main_iteration()
                    if self.FileTransferCancelled:
                        raise UserCancelEvent
                    if (result == SM_FAIL):
                        raise RemoteExecError
                else:
                    total_made = 0
                    for dir in dirs:
                        self.UpdateFileTransferProgressWindow(dir, float(total_made)/float(total_dirs))
                        result, output = self.MakeRemoteDirectory(machine, dir)
                        if self.FileTransferCancelled:
                            raise UserCancelEvent
                        if (result == SM_FAIL):
                            raise RemoteExecError
                        total_made += 1
                #
                # Transfer files to the remote machine.
                #
                self.ProgressWindow.name.set_text("Transferring files...")
                self.Data.LogMessage("Transferring files...")
                self.TotalFilesToSend = len(paths)
                if (result != SM_FAIL) and (self.TotalFilesToSend > 0):
                    self.TotalFilesSent = 0
                    for path in paths:
                        rpathlist = [dest]
                        rpathlist.extend(self.Data.PathList(path))
                        rpath  = self.Data.RemotePath(rpathlist, machine.Name)
                        result = self.TransferFiles(SM_LOCALHOST, [path], machine, rpath, False, None)
                        if self.FileTransferCancelled:
                            raise UserCancelEvent
                        if (result == SM_FAIL):
                            raise RemoteExecError
                self.Data.LogMessage("Source code transfer complete.")
            except UserCancelEvent:
                self.Data.LogMessage("File transfer cancelled by user.")
            except RemoteExecError:
                self.Data.LogMessage("Error while uploading source files.")
            except IOError:
                self.Data.LogMessage("I/O error while uploading source files.")
                self.NotifyDialog("I/O error!  Perhaps a file or directory has been removed?")
            self.CloseFileTransferProgressWindow()
            os.chdir(save_cwd)
        return True
        
    def BrowseForConfigFile (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        result, filename = self.FileChooserDialog(self.Data.CurrentExperiment.ConfigFile, \
                                                  "FileChooserDialog")
        if (result == gtk.RESPONSE_OK):
            self.Data.CurrentExperiment.ConfigFile = filename
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        
    def UploadConfigFile (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        machine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        if (not machine):
            self.NotifyDialog("Please choose an execution machine before attempting to upload a configuration file.")
            return True
        app     = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        if (not app):
            self.NotifyDialog("Please choose an application before attempting to upload a configuration file.")
            return True
        file    = self.Data.CurrentExperiment.ConfigFile
        dest    = self.Data.RemotePath([self.Data.CurrentExperiment.RemoteBuildDir, app.ConfigFileName], \
                                       machine.Name)
        result, output = self.MakeRemoteDirectory(machine, self.Data.CurrentExperiment.RemoteBuildDir)
        if (result == SM_OK):
            self.TransferFiles(SM_LOCALHOST, [file], machine, dest, True, self.wTree.window)
        return True
    
    def NotifyDialog (self, text):
        dTree = gtk.glade.XML(GladeFile, "NotifyDialog")
        dTree.dialog = dTree.get_widget("NotifyDialog")
        dTree.get_widget("label253").set_text(text)
        dTree.dialog.present()
        dTree.dialog.run()
        dTree.dialog.destroy()
    
    def QuestionDialog (self, text):
        dTree = gtk.glade.XML(GladeFile, "QuestionDialog")
        dTree.dialog = dTree.get_widget("QuestionDialog")
        dTree.get_widget("label255").set_text(text)
        dTree.dialog.present()
        result = dTree.dialog.run()
        dTree.dialog.destroy()
        return result
    
    def QuitDialog (self):
        dTree = gtk.glade.XML(GladeFile, "QuitDialog")
        dTree.dialog = dTree.get_widget("QuitDialog")
        dTree.dialog.present()
        result = dTree.dialog.run()
        dTree.dialog.destroy()
        return result
        
    def SaveQuestionDialog (self, text, dialog="SaveQuestionDialog", label="label279"):
        dTree = gtk.glade.XML(GladeFile, dialog)
        dTree.dialog = dTree.get_widget(dialog)
        dTree.get_widget(label).set_text(text)
        dTree.dialog.present()
        result = dTree.dialog.run()
        dTree.dialog.destroy()
        return result

    def ConfigureApp (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        if not app:
            self.NotifyDialog("Please choose an application before attempting to configure.")
            return True
        machine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        if not machine:
            self.NotifyDialog("Please choose an execution machine before configuring your application.")
            return True
        command  = self.Data.CurrentExperiment.ConfigCommand
        startdir = self.Data.CurrentExperiment.RemoteBuildDir
        self.OpenRemoteCommandProgressWindow("Configuring application", command, self.wTree.window)
        result, output = self.RemoteExec(machine, startdir, command)
        self.CloseRemoteCommandProgressWindow()
        self.Data.CurrentExperiment.ConfigStatus = result
        self.Data.CurrentExperiment.ConfigOutput = output
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        if (self.Data.CurrentExperiment.ConfigStatus == SM_OK):
            result = self.ExperimentDialog.ConfigButton.modify_bg(gtk.STATE_NORMAL, \
                                                                  gtk.gdk.color_parse("green"))
        elif (self.Data.CurrentExperiment.ConfigStatus == SM_FAIL):
            result = self.ExperimentDialog.ConfigButton.modify_bg(gtk.STATE_NORMAL, \
                                                                  gtk.gdk.color_parse("red"))
        return True
        
    def BuildApp (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        app = self.Data.Applications.EntryByName(self.Data.CurrentExperiment.Application)
        if not app:
            self.NotifyDialog("Please choose an application before attempting to build.")
            return True
        machine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        if not machine:
            self.NotifyDialog("Please choose an execution machine before building your application.")
            return True
        if app.MoveExecutable:
            result, output = self.MakeRemoteDirectory(machine, machine.RemoteExecDir)
        if (result == SM_OK):
            startdir = self.Data.CurrentExperiment.RemoteBuildDir
            command  = self.Data.CurrentExperiment.BuildCommand
            self.OpenRemoteCommandProgressWindow("Building application", command, self.wTree.window)
            result, output = self.RemoteExec(machine, startdir, command)
            self.CloseRemoteCommandProgressWindow()
            self.Data.CurrentExperiment.BuildStatus = result
            self.Data.CurrentExperiment.BuildOutput = output
            self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        if (self.Data.CurrentExperiment.BuildStatus == SM_OK):
            result = self.ExperimentDialog.BuildButton.modify_bg(gtk.STATE_NORMAL, \
                                                                 gtk.gdk.color_parse("green"))
        elif (self.Data.CurrentExperiment.BuildStatus == SM_FAIL):
            result = self.ExperimentDialog.BuildButton.modify_bg(gtk.STATE_NORMAL, \
                                                                 gtk.gdk.color_parse("red"))
        return True

    def BrowseForParmTemplate (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        result, filename = self.FileChooserDialog(self.Data.CurrentExperiment.ParmFileTemplate, \
                                                  "FileChooserDialog")
        if (result == gtk.RESPONSE_OK):
            self.Data.CurrentExperiment.ParmFileTemplate = filename
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        return True

    def EditParmTemplate (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        self.EditTextFile(self.Data.CurrentExperiment.ParmFileTemplate)
        self.SetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)
        return True

    def CreateDialog (self, name, fields, signals=None, listname=None):
        dTree = gtk.glade.XML(GladeFile, name)
        dTree.dialog = dTree.get_widget(name)
        dTree.dialog.hide()
        if listname:
            dTree.treestore = gtk.ListStore(str, object)
            dTree.treeview  = dTree.get_widget(listname)
            dTree.treeview.set_model(dTree.treestore)
            dTree.tvcolumn  = gtk.TreeViewColumn("Name")
            dTree.treeview.append_column(dTree.tvcolumn)
            dTree.cell      = gtk.CellRendererText()
            dTree.tvcolumn.pack_start(dTree.cell, True)
            dTree.tvcolumn.add_attribute(dTree.cell, "text", 0)
        dTree.handler_ids = { }
        if signals:
            for widget_name in signals.keys():
                signal = signals[widget_name]
                if (len(signal) == 2):
                    widget = dTree.get_widget(widget_name)
                else:
                    parent = dTree.get_widget(widget_name)
                    widget = apply(getattr(parent, signal[2]))
                id     = widget.connect(signal[0], signal[1])
                dTree.handler_ids[widget_name] = [widget, id]
        dTree.fields = fields
        dTree.dialog.set_transient_for(self.wTree.window)
        return dTree

    def SetDialogData (self, dTree, object, choices=None):
        for widget_name in dTree.handler_ids.keys():
            widget = dTree.handler_ids[widget_name][0]
            widget.handler_block(dTree.handler_ids[widget_name][1])
        for field in dTree.fields.keys():
            widget    = dTree.get_widget(field)
            attribute = dTree.fields[field]
            if hasattr(object, attribute):
                if isinstance(widget, gtk.SpinButton):
                    widget.set_value(getattr(object, attribute))
                elif isinstance(widget, gtk.FontButton):
                    widget.set_font_name(getattr(object, attribute))
                elif isinstance(widget, gtk.ColorButton):
                    widget.set_color(getattr(object, attribute))
                elif isinstance(widget, gtk.Entry) or isinstance(widget, gtk.Label):
                    widget.set_text(getattr(object, attribute))
                elif isinstance(widget, gtk.TextView):
                    buffer = widget.get_buffer()
                    buffer.set_text(getattr(object, attribute))
                elif isinstance(widget, gtk.RadioButton):
                    group = widget.get_group()
                    group[getattr(object,attribute)].set_active(True)
                elif isinstance(widget, gtk.Button):
                    widget.set_active(getattr(object, attribute))
                elif isinstance(widget, gtk.ComboBoxEntry):
                    if choices:
                        if (attribute in choices.keys()):
                            treemodel = widget.get_model()
                            treeiter  = treemodel.get_iter_first()
                            while treeiter:
                                treemodel.remove(treeiter)
                                treeiter = treemodel.get_iter_first()
                            for choice in choices[attribute]:
                                widget.append_text(choice)
                    entry = widget.child
                    entry.set_text(getattr(object, attribute))
                elif isinstance(widget, gtk.ComboBox):
                    if choices:
                        if (attribute in choices.keys()):
                            treemodel = widget.get_model()
                            treeiter  = treemodel.get_iter_first()
                            while treeiter:
                                treemodel.remove(treeiter)
                                treeiter = treemodel.get_iter_first()
                            for choice in choices[attribute]:
                                widget.append_text(choice)
                    treemodel = widget.get_model()
                    index     = -1
                    iter      = treemodel.get_iter_first()
                    i         = 0
                    while iter:
                        value = treemodel.get_value(iter, 0)
                        if (value == getattr(object, attribute)):
                            index = i
                        i = i + 1
                        iter = treemodel.iter_next(iter)
                    if (index >= 0):
                        widget.set_active(index)
            else:
                self.Data.LogMessage("SetDialogData: object %s has no attribute %s" % (object, attribute))
        for widget_name in dTree.handler_ids.keys():
            widget = dTree.handler_ids[widget_name][0]
            widget.handler_unblock(dTree.handler_ids[widget_name][1])

    def GetDialogData (self, dTree, object, clear=False):
        for widget_name in dTree.handler_ids.keys():
            widget = dTree.handler_ids[widget_name][0]
            widget.handler_block(dTree.handler_ids[widget_name][1])
        for field in dTree.fields.keys():
            widget    = dTree.get_widget(field)
            attribute = dTree.fields[field]
            if hasattr(object, attribute):
                if isinstance(widget, gtk.SpinButton):
                    setattr(object, attribute, widget.get_value_as_int())
                elif isinstance(widget, gtk.FontButton):
                    setattr(object, attribute, widget.get_font_name())
                elif isinstance(widget, gtk.ColorButton):
                    setattr(object, attribute, widget.get_color())
                elif isinstance(widget, gtk.Entry) or isinstance(widget, gtk.Label):
                    setattr(object, attribute, widget.get_text())
                    if clear:
                        widget.set_text("")
                elif isinstance(widget, gtk.TextView):
                    buffer = widget.get_buffer()
                    setattr(object, attribute, buffer.get_text(buffer.get_start_iter(), \
                                                               buffer.get_end_iter()))
                    if clear:
                        buffer.set_text("")
                elif isinstance(widget, gtk.RadioButton):
                    group = widget.get_group()
                    i = 0
                    for member in group:
                        if member.get_active():
                            setattr(object, attribute, i)
                            break
                        else:
                            i = i + 1
                elif isinstance(widget, gtk.Button):
                    setattr(object, attribute, widget.get_active())
                elif isinstance(widget, gtk.ComboBoxEntry):
                    entry = widget.child
                    setattr(object, attribute, entry.get_text())
                    if clear:
                        entry.set_text("")
                elif isinstance(widget, gtk.ComboBox):
                    treemodel = widget.get_model()
                    index     = widget.get_active()
                    if index >= 0:
                        setattr(object, attribute, treemodel[index][0])
            else:
                self.Data.LogMessage("GetDialogData: object %s has no attribute %s" % (object, attribute))
        if hasattr(object, "ModifiedDate"):
            setattr(object, "ModifiedDate", time.asctime(time.localtime()))
        self.Data.Changed = True
        for widget_name in dTree.handler_ids.keys():
            widget = dTree.handler_ids[widget_name][0]
            widget.handler_unblock(dTree.handler_ids[widget_name][1])

    def BrowseForWorkspace (self, widget, data=None):
        self.GetDialogData(self.PreferencesDialog, self.Data.CurrentUser)
        result, filename = self.FileChooserDialog(self.Data.CurrentUser.Workspace, \
                                                  "FolderChooserDialog")
        if (result == gtk.RESPONSE_OK):
            self.Data.CurrentUser.Workspace = filename
        self.SetDialogData(self.PreferencesDialog, self.Data.CurrentUser)
        return True

    def SwitchGridProxyMechanism (self, widget, data=None):
        value = self.PreferencesDialog.get_widget("checkbutton17").get_active()
        if value:
            self.PreferencesDialog.get_widget("label310").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry107").set_sensitive(True)
            self.PreferencesDialog.get_widget("label320").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry109").set_sensitive(True)
            self.PreferencesDialog.get_widget("label307").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry102").set_sensitive(False)
        else:
            self.PreferencesDialog.get_widget("label310").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry107").set_sensitive(False)
            self.PreferencesDialog.get_widget("label320").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry109").set_sensitive(False)
            self.PreferencesDialog.get_widget("label307").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry102").set_sensitive(True)
        return True
        
    def EditUserData (self, widget=None, data=None):
        self.ConvertUserColors(True)
        self.SetDialogData(self.PreferencesDialog, self.Data.CurrentUser)
        self.ConvertUserColors(False)
        if self.Data.CurrentUser.UseMyProxy:
            self.PreferencesDialog.get_widget("label310").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry107").set_sensitive(True)
            self.PreferencesDialog.get_widget("label320").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry109").set_sensitive(True)
            self.PreferencesDialog.get_widget("label307").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry102").set_sensitive(False)
        else:
            self.PreferencesDialog.get_widget("label310").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry107").set_sensitive(False)
            self.PreferencesDialog.get_widget("label320").set_sensitive(False)
            self.PreferencesDialog.get_widget("entry109").set_sensitive(False)
            self.PreferencesDialog.get_widget("label307").set_sensitive(True)
            self.PreferencesDialog.get_widget("entry102").set_sensitive(True)
        result = None
        while ((result != gtk.RESPONSE_CANCEL) & (result != gtk.RESPONSE_OK) &
               (result != gtk.RESPONSE_DELETE_EVENT)):
            Backup = copy.deepcopy(self.Data.CurrentUser)
            result = self.PreferencesDialog.dialog.run()
            if (result != gtk.RESPONSE_CANCEL):
                self.GetDialogData(self.PreferencesDialog, self.Data.CurrentUser)
                self.ConvertUserColors(False)
                self.UpdateAllTreeStoreRows()
                self.Data.CurrentUser.DataEntered = True
            else:
                self.Data.CurrentUser.__dict__ = Backup.__dict__
                self.UpdateAllTreeStoreRows()
        self.PreferencesDialog.dialog.hide()
        self.ApplyUserSettings()
        return True

    def ApplyUserSettings (self):
        gobject.source_remove(self.UpdateTimeout)
        if self.Data.CurrentUser.AutoUpdate:
            self.UpdateTimeout = gobject.timeout_add(self.Data.CurrentUser.AutoUpdateInterval*60000, \
                                                     self.UpdateAllJobStatus, None)
    
    def ConvertUserColors (self, TuplesToColors):
        user = self.Data.CurrentUser
        if TuplesToColors:
            user.ProjectTextColor    = self.TupleToColor(user.ProjectTextColor)
            user.ExperimentTextColor = self.TupleToColor(user.ExperimentTextColor)
            user.RunTextColor        = self.TupleToColor(user.RunTextColor)
            user.JobTextColor        = self.TupleToColor(user.JobTextColor)
        else:
            user.ProjectTextColor    = self.ColorToTuple(user.ProjectTextColor)
            user.ExperimentTextColor = self.ColorToTuple(user.ExperimentTextColor)
            user.RunTextColor        = self.ColorToTuple(user.RunTextColor)
            user.JobTextColor        = self.ColorToTuple(user.JobTextColor)
    
    def TupleToColor (self, tuple):
        return gtk.gdk.Color(red=tuple[0], green=tuple[1], blue=tuple[2])
        
    def ColorToTuple (self, color):
        return (color.red, color.green, color.blue)
        
    def EditApplicationData (self, widget, data=None):
        self.SetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
        self.PopulateResListTreeView(self.ApplicationDialog, self.Data.Applications)
        result = None
        while ((result != gtk.RESPONSE_CANCEL) & (result != gtk.RESPONSE_OK) &
               (result != gtk.RESPONSE_DELETE_EVENT)):
            Backup = copy.deepcopy(self.Data.Applications)
            result = self.ApplicationDialog.dialog.run()
            if (result != gtk.RESPONSE_CANCEL):
                self.GetDialogData(self.ApplicationDialog, self.Data.Applications.Current)
                if self.Data.Applications.Current in self.Data.Applications.Entries:
                    self.UpdateResListTreeStoreRow(self.ApplicationDialog, \
                                                   self.Data.Applications)
            else:
                self.Data.Applications.__dict__ = Backup.__dict__
        self.ApplicationDialog.dialog.hide()
        return True
    
    def SetMachineDialogData (self):
        self.SetDialogData(self.MachineDialog, self.Data.Machines.Current, \
                           { "AccessMethod"   : self.Data.AccessNames, \
                             "QueueSystem"    : self.Data.QueueNames, \
                             "DefaultQueue"   : self.Data.Machines.Current.Queues, \
                             "DefaultAccount" : self.Data.Machines.Current.Accounts })
        if (self.Data.Machines.Current.AccessMethodType == SM_AUTH_KERBEROS):
            self.MachineDialog.get_widget("label286").set_sensitive(True)
            self.MachineDialog.get_widget("entry92").set_sensitive(True)
        else:
            self.MachineDialog.get_widget("label286").set_sensitive(False)
            self.MachineDialog.get_widget("entry92").set_sensitive(False)
        if (self.Data.Machines.Current.AccessMethodType == SM_AUTH_GSI):
            self.MachineDialog.get_widget("label324").set_sensitive(True)
            self.MachineDialog.get_widget("entry111").set_sensitive(True)
            self.MachineDialog.get_widget("label325").set_sensitive(True)
            self.MachineDialog.get_widget("entry112").set_sensitive(True)
        else:
            self.MachineDialog.get_widget("label324").set_sensitive(False)
            self.MachineDialog.get_widget("entry111").set_sensitive(False)
            self.MachineDialog.get_widget("label325").set_sensitive(False)
            self.MachineDialog.get_widget("entry112").set_sensitive(False)
        
    def EditMachineData (self, widget, data=None):
        self.SetMachineDialogData()
        self.PopulateResListTreeView(self.MachineDialog, self.Data.Machines)
        result = None
        while ((result != gtk.RESPONSE_CANCEL) & (result != gtk.RESPONSE_OK) &
               (result != gtk.RESPONSE_DELETE_EVENT)):
            Backup = copy.deepcopy(self.Data.Machines)
            result = self.MachineDialog.dialog.run()
            if (result != gtk.RESPONSE_CANCEL):
                self.GetDialogData(self.MachineDialog, self.Data.Machines.Current)
                if self.Data.Machines.Current in self.Data.Machines.Entries:
                    self.UpdateResListTreeStoreRow(self.MachineDialog, \
                                                   self.Data.Machines)
            else:
                self.Data.Machines.__dict__ = Backup.__dict__
        self.MachineDialog.dialog.hide()
        return True
    
    def UpdateResListTreeStoreRow (self, dialog, reslist):
        iter = dialog.treestore.get_iter_first()
        while iter:
            if (dialog.treestore.get_value(iter, 1) == reslist.Current):
                dialog.treestore.set(iter, 0, reslist.Current.Name)
            iter = dialog.treestore.iter_next(iter)
        
    def AddProject (self, widget, data=None):
        self.Data.AddProject()
        self.AddTreeStoreRow(None, self.Data.CurrentProject.Name, \
                                   self.Data.CurrentProject)
        self.EditProjectData()
        return True

    def TreeViewRowStyle (self, container):
        if isinstance(container, Run):
            Prefix     = "Run "
            TextColor  = self.TupleToColor(self.Data.CurrentUser.RunTextColor)
            TextFont   = pango.FontDescription(self.Data.CurrentUser.RunTextFont)
        elif isinstance(container, Job):
            Prefix     = "Job "
            TextColor  = self.TupleToColor(self.Data.CurrentUser.JobTextColor)
            TextFont   = pango.FontDescription(self.Data.CurrentUser.JobTextFont)
        elif isinstance(container, Project):
            Prefix     = ""
            TextColor  = self.TupleToColor(self.Data.CurrentUser.ProjectTextColor)
            TextFont   = pango.FontDescription(self.Data.CurrentUser.ProjectTextFont)
        else:
            Prefix     = ""
            TextColor  = self.TupleToColor(self.Data.CurrentUser.ExperimentTextColor)
            TextFont   = pango.FontDescription(self.Data.CurrentUser.ExperimentTextFont)
        return (Prefix, TextColor, TextFont)

    def AddTreeStoreRow (self, parent, name, container):
        Prefix, TextColor, TextFont = self.TreeViewRowStyle(container)
        ParentIter = self.FindTreeIter(self.treestore.iter_children(None), \
                                       parent)
        treeiter = self.treestore.append(ParentIter, [Prefix + name, container, \
                                                      container.Comment, \
                                                      container.StatusLine(), \
                                                      container.ModifiedDate, \
                                                      TextColor, \
                                                      True, \
                                                      TextFont])
        if (len(container.Children) > 0):
            for child in container.Children:
                self.AddTreeStoreRow(container, child.Name, child)
        treeselection = self.treeview.get_selection()
        treeselection.unselect_all()
        treeselection.select_iter(treeiter)
        
    def FindTreeIter (self, iter, container):
        while iter:
            if (self.treestore.get_value(iter, 1) == container):
                return iter
            result = self.FindTreeIter(self.treestore.iter_children(iter), \
                                       container)
            if result: return result
            iter = self.treestore.iter_next(iter)
        return None	

    def UpdateTreeStoreRow (self, container):
        treeiter = self.FindTreeIter(self.treestore.iter_children(None), \
                                     container)
        Prefix, TextColor, TextFont = self.TreeViewRowStyle(container)
        self.treestore.set(treeiter, 0, Prefix + container.Name)
        self.treestore.set(treeiter, 2, container.Comment)
        self.treestore.set(treeiter, 3, container.StatusLine())
        self.treestore.set(treeiter, 4, container.ModifiedDate)
        self.treestore.set(treeiter, 5, TextColor)
        self.treestore.set(treeiter, 6, True)
        self.treestore.set(treeiter, 7, TextFont)

    def UpdateAllTreeStoreRows (self, container=None):
        if container:
            self.UpdateTreeStoreRow(container)
            for child in container.Children:
                self.UpdateAllTreeStoreRows(child)
        else:
            for project in self.Data.Projects:
                self.UpdateAllTreeStoreRows(project)
    
    def GetTreePaths (self, iter, pathlist):
        while iter:
            pathlist.append(self.treestore.get_path(iter))
            self.GetTreePaths(self.treestore.iter_children(iter), pathlist)
            iter = self.treestore.iter_next(iter)
    
    def RemoveTreeStoreRow (self, iter):
        pathlist = [self.treestore.get_path(iter)]
        self.GetTreePaths(self.treestore.iter_children(iter), pathlist)
        pathlist.reverse()
        for path in pathlist:
            self.treestore.remove(self.treestore.get_iter(path))
    
    def AddExperiment (self, widget, data=None):
        self.Data.AddExperiment()
        self.AddTreeStoreRow(self.Data.CurrentExperiment.Parent, \
                             self.Data.CurrentExperiment.Name, \
                             self.Data.CurrentExperiment)
        self.EditExperimentData()
        return True

    def AddJob (self, widget, data=None):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.RunDialog, self.Data.CurrentRun)
        machine = self.Data.Machines.EntryByName(self.Data.CurrentRun.ExecMachine)
        if not machine:
            self.NotifyDialog("Please choose an execution machine before creating a job.")
            return True
        app = self.Data.Applications.EntryByName(self.Data.CurrentRun.Application)
        if not app:
            self.NotifyDialog("Please choose an application before creating a job.")
            return True
        self.Data.CurrentJob = self.Data.CurrentRun.AddChild(self.Data.CurrentUser)
        self.AddTreeStoreRow(self.Data.CurrentRun, self.Data.CurrentJob.Name, self.Data.CurrentJob)
        self.EditJobData()
        return True
        
    def AttachRun (self, widget, data=None):
        pass

    def ViewRunParameters (self, widget, data=None):
        self.ViewWindow("Parameter file for run " + self.Data.CurrentRun.Name, \
                        self.Data.CurrentRun.ParmFileData, \
                        use_dialog=self.FileViewWindow)
        
    def ViewJobParameters (self, widget, data=None):
        self.ViewWindow("Parameter file for job " + self.Data.CurrentJob.Name, \
                        self.Data.CurrentJob.ParmFileData, \
                        use_dialog=self.FileViewWindow)
    
    def GetJobLog (self, job): 
        if job.LogFileFinal:
            return SM_OK
        try:
            app = self.Data.Applications.EntryByName(job.Application)
            if not app:
                self.NotifyDialog("No application chosen for this job.")
                return SM_FAIL
            if (app.LogFileName != ""):
                machine = self.Data.Machines.EntryByName(job.ExecMachine)
                if not machine:
                    self.NotifyDialog("No execution machine chosen for this job.")
                    return SM_FAIL
                result = self.TransferFiles(machine, [self.Data.RemotePath([job.RemoteRunDir, app.LogFileName],\
                                                                           machine.Name)], \
                                            SM_LOCALHOST, TempFile, False, None)
                if (result == SM_OK):
                    f = open(TempFile)
                    job.LogFileData = f.read()
                    f.close()
                    if (job.RunStatus == SM_STATUS_DONE):
                        job.LogFileFinal = True
        except IOError:
            self.Notify("Error writing local temp file.")
            return SM_FAIL
        return result

    def ViewJobLog (self, widget, data=None):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        self.LockUpdates()
        self.GetJobStatus(widget)
        self.UnlockUpdates()
        if (self.Data.CurrentJob.RunStatus != SM_STATUS_NOTAVAIL):
            self.LockUpdates()
            result = self.GetJobLog(self.Data.CurrentJob)
            self.UnlockUpdates()
            if (result == SM_OK):
                text = self.Data.CurrentJob.LogFileData
                self.ViewWindow("Log file for job " + self.Data.CurrentJob.Name, text, \
                                use_dialog=self.FileViewWindow)
        if isinstance(widget, gtk.Button):
            self.SetDialogData(self.JobDialog, self.Data.CurrentJob)
        return True

    def GetJobOutput (self, job):
        result = SM_OK
        try:
            machine = self.Data.Machines.EntryByName(job.ExecMachine)
            if not machine:
                self.NotifyDialog("No execution machine chosen for this job.")
                return SM_FAIL
            if not job.StdoutFinal:
                result = self.TransferFiles(machine, [job.RemoteStdoutFile], \
                                            SM_LOCALHOST, TempFile, False, None)
                if (result == SM_OK):
                    f = open(TempFile)
                    job.StdoutData = f.read()
                    f.close()
                    if (job.RunStatus == SM_STATUS_DONE):
                        job.StdoutFinal = True
            if not job.StderrFinal:
                result = self.TransferFiles(machine, [job.RemoteStderrFile], \
                                            SM_LOCALHOST, TempFile, False, None)
                if (result == SM_OK):
                    f = open(TempFile)
                    job.StderrData = f.read()
                    f.close()
                    if (job.RunStatus == SM_STATUS_DONE):
                        job.StderrFinal = True
        except IOError:
            self.Notify("Error writing local temp file.")
            return SM_FAIL
        return result
    
    def ViewJobOutput (self, widget, data=None):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        self.LockUpdates()
        self.GetJobStatus(widget)
        self.UnlockUpdates()
        if (self.Data.CurrentJob.RunStatus != SM_STATUS_NOTAVAIL):
            self.LockUpdates()
            result = self.GetJobOutput(self.Data.CurrentJob)
            self.UnlockUpdates()
            if (result == SM_OK):
                text = "===== Standard output =====\n" + \
                       self.Data.CurrentJob.StdoutData + "\n\n" + \
                       "===== Standard error =====\n" + \
                       self.Data.CurrentJob.StderrData
                self.ViewWindow("Output for job " + self.Data.CurrentJob.Name, text, \
                                use_dialog=self.FileViewWindow)
        if isinstance(widget, gtk.Button):
            self.SetDialogData(self.JobDialog, self.Data.CurrentJob)
        return True
    
    def ViewConfigOutput (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)    # just to be consistent w/behavior of other buttons
        if (self.Data.CurrentExperiment.ConfigStatus != SM_UNKNOWN):
            self.ViewWindow("Remote configuration output", self.Data.CurrentExperiment.ConfigOutput, \
                            use_dialog=self.FileViewWindow)
        return True
    
    def ViewBuildOutput (self, widget, data=None):
        self.GetDialogData(self.ExperimentDialog, self.Data.CurrentExperiment)    # just to be consistent w/behavior of other buttons
        if (self.Data.CurrentExperiment.BuildStatus != SM_UNKNOWN):
            self.ViewWindow("Remote build output", self.Data.CurrentExperiment.BuildOutput, \
                            use_dialog=self.FileViewWindow)
        return True

    def ViewWindow (self, title, text, use_dialog=None):
        Output = ParsableString(text)
        if use_dialog:
            dialog = use_dialog
        else:
            dialog = self.OutputWindow
        self.SetDialogData(dialog, Output)
        dialog.dialog.set_title(title)
        dialog.dialog.run()
        dialog.dialog.hide()
        return True

    def ExtendTendril (self, machine):
        if not os.path.exists(TendrilPath):
            self.NotifyDialog("Could not access tendril executable at:\n\n"+TendrilPath+"\n\n"+\
                              "Please disable tendril in your user settings.")
            return SM_FAIL, ""
        stat_tuple = os.stat(TendrilPath)
        mode       = stat_tuple[stat.ST_MODE]
        if (mode & stat.S_IWGRP) or (mode & stat.S_IWOTH):     # make sure only we and root can write to tendril.py
            self.NotifyDialog("Insecure permissions on the tendril executable at:\n\n"+TendrilPath+"\n\n"+\
                              "Please disable tendril in your user settings.")
            return SM_FAIL, ""
        if machine.TeuthisDir:
            result, output = self.MakeRemoteDirectory(machine, machine.TeuthisDir)
            RemoteTendrilPath = self.Data.RemotePath([machine.TeuthisDir, "tendril.py"], machine.Name)
        else:
            result = SM_OK
            RemoteTendrilPath = "tendril.py"
        if (result == SM_OK):
            result = self.TransferFiles(SM_LOCALHOST, [TendrilPath], machine, machine.TeuthisDir, False, None, True)
        return result, RemoteTendrilPath

    def StageJobData (self, job):
        DataSrcMachine = self.Data.Machines.EntryByName(job.DataSrcMachine)
        ExecMachine    = self.Data.Machines.EntryByName(job.ExecMachine)
        if (not DataSrcMachine) or (not ExecMachine):
            self.NotifyDialog("Please specify data source and execution machines" + \
                              " before staging files.")
            return SM_FAIL
        self.Data.LogMessage("Staging job data...")
        result, output = self.MakeRemoteDirectory(ExecMachine, job.RemoteRunDir)
        if (result == SM_OK):
            DataSrcFileList = job.DataSrcFiles.split("\n")
            result = self.TransferFiles(DataSrcMachine, DataSrcFileList, ExecMachine, job.RemoteRunDir, \
                                        True, self.wTree.window, DestIsDir=True)
            return result
        else:
            self.Data.LogMessage("Error creating run directory on execution machine.")
            self.NotifyDialog("Error creating run directory on execution machine.")
            return SM_FAIL

    def ArchiveJobDataFromButton (self, widget, data=None):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        self.GetJobStatus(widget, ArchivePrompt=False)
        if (self.Data.CurrentJob.RunStatus != SM_STATUS_NOTAVAIL):
            self.GetJobLog(self.Data.CurrentJob)
            self.GetJobOutput(self.Data.CurrentJob)
            self.ArchiveJobData(self.Data.CurrentJob)
        if isinstance(widget, gtk.Button):
            self.SetDialogData(self.JobDialog, self.Data.CurrentJob)
        return True
    
    def ArchiveJobData (self, job):
        DataDestMachine = self.Data.Machines.EntryByName(job.DataDestMachine)
        ExecMachine     = self.Data.Machines.EntryByName(job.ExecMachine)
        if (not DataDestMachine) or (not ExecMachine):
            self.NotifyDialog("Please specify execution and data destination machines" + \
                              " before archiving output files.")
            return SM_FAIL
        self.Data.LogMessage("Archiving job data...")
        result, output = self.MakeRemoteDirectory(DataDestMachine, job.DataDestPath)
        if (result == SM_OK):
            if self.Data.CurrentUser.UseTendril:
                result, RemoteTendrilPath = self.ExtendTendril(ExecMachine)
                if (result == SM_OK):
                    result, output = self.RemoteExec(ExecMachine, "", RemoteTendrilPath+" listfiles "+job.RemoteRunDir)
                    source_files = output.split("\n")
                    if (result == SM_OK):
                        result = self.TransferFiles(ExecMachine, source_files, DataDestMachine, job.DataDestPath, \
                                                    True, self.wTree.window, DestIsDir=True)
            else:
                all_files    = OSTypes[ExecMachine.OSType]["AllFilesWildcard"]
                source_files = self.Data.RemotePath([job.RemoteRunDir, all_files], ExecMachine.Name)
                result = self.TransferFiles(ExecMachine, [source_files], DataDestMachine, job.DataDestPath, \
                                            True, self.wTree.window, DestIsDir=True)
            return result
        else:
            self.Data.LogMessage("Error creating path on data destination machine.")
            self.NotifyDialog("Error creating path on data destination machine.")
            return SM_FAIL

    def SubmitJob (self, widget, data=None):
        #
        # Get the job data and make sure everything is OK
        #
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        if (self.Data.CurrentJob.RunStatus in [SM_STATUS_PEND, SM_STATUS_RUN, SM_STATUS_DONE]):
            self.NotifyDialog("Job has already been submitted. If you would like to resubmit, try cloning the job.")
            return True
        machine = self.Data.Machines.EntryByName(self.Data.CurrentJob.ExecMachine)
        if not machine:
            self.NotifyDialog("No execution machine is defined for this job.")
            return True
        system  = self.Data.QueuingSystemByName(machine.QueueSystem)
        if not system:
            self.NotifyDialog("The execution machine for this job has no queuing system configured for it.")
            return True
        app = self.Data.Applications.EntryByName(self.Data.CurrentJob.Application)
        if not app:
            self.NotifyDialog("No application is defined for this job.")
            return True
        #
        # Generate the job script and let the user check it over
        #
        self.Data.CurrentJob.GenerateJobScript(self.Data)
        Output = ParsableString(self.Data.CurrentJob.JobScript)
        self.SetDialogData(self.SubmitDialog, Output)
        result = self.SubmitDialog.dialog.run()
        self.SubmitDialog.dialog.hide()
        #
        if (result == gtk.RESPONSE_OK):
            try:
                #
                # Stage job data
                #
                data_src = self.Data.Machines.EntryByName(self.Data.CurrentJob.DataSrcMachine)
                if (data_src and (self.Data.CurrentJob.DataSrcFiles != "")):
                    result = self.StageJobData(self.Data.CurrentJob)
                    if (result != SM_OK):
                        return True
                #
                # Get the job script and send it to the execution machine
                #
                self.GetDialogData(self.SubmitDialog, Output)
                self.Data.CurrentJob.JobScript = Output.String
                f = open(TempFile, 'w')
                f.write(self.Data.CurrentJob.JobScript)
                f.close()
                if machine.TeuthisDir:
                    result, output = self.MakeRemoteDirectory(machine, machine.TeuthisDir)
                    if (result != SM_OK):
                        return True
                    self.Data.CurrentJob.RemoteScriptFile = self.Data.RemotePath([machine.TeuthisDir, RemoteScriptFile], \
                                                                                 machine.Name)
                else:
                    self.Data.CurrentJob.RemoteScriptFile = RemoteScriptFile
                result = self.TransferFiles(SM_LOCALHOST, [TempFile], machine, \
                                            self.Data.CurrentJob.RemoteScriptFile, \
                                            True, self.wTree.window)
                if (result != SM_OK):
                    return True
                #
                # Create the remote run directory
                #
                result, output = self.MakeRemoteDirectory(machine, self.Data.CurrentJob.RemoteRunDir)
                if (result != SM_OK):
                    return True
                #
                # Send the runtime parameter file
                #
                if (app.ParamFileName != ""):
                    f = open(TempFile, 'w')
                    f.write(self.Data.CurrentJob.ParmFileData)
                    f.close()
                    result = self.TransferFiles(SM_LOCALHOST, [TempFile], machine, \
                                                self.Data.RemotePath([self.Data.CurrentJob.RemoteRunDir, app.ParamFileName], \
                                                                     machine.Name), \
                                                True, self.wTree.window)
                    if (result != SM_OK):
                        return True
                #
                # Submit the job
                #
                command = system.GetSubmitCommand(self.Data.CurrentJob)
                result, output = self.RemoteExec(machine, machine.TeuthisDir, command)
                if (result == SM_OK):
                    self.Data.CurrentJob.RemoteJobID      = system.GetJobID(output)
                    self.Data.CurrentJob.RemoteStdoutFile = system.GetRemoteStdoutFileName(self.Data.CurrentJob, machine)
                    self.Data.CurrentJob.RemoteStderrFile = system.GetRemoteStderrFileName(self.Data.CurrentJob, machine)
                    self.Data.CurrentJob.SubmitDate       = time.asctime(time.localtime())
                    self.Data.CurrentJob.RunStatus        = SM_STATUS_PEND
                    self.UpdateTreeStoreRow(self.Data.CurrentJob)
                    self.Data.LogMessage("Submitted job %s at %s" % (self.Data.CurrentJob.RemoteJobID, self.Data.CurrentJob.SubmitDate))
                else:
                    self.NotifyDialog("Error submitting job; job not submitted.")
                    return True
            except IOError:
                self.NotifyDialog("Error writing temporary file; job not submitted.")
            if isinstance(widget, gtk.Button):
                self.SetDialogData(self.JobDialog, self.Data.CurrentJob)
        return True
    
    def LockUpdates (self):
        self.UpdateLock = True
        
    def UnlockUpdates (self):
        self.UpdateLock = False
        
    def UpdateAllJobStatus (self, widget, data=None):
        # use update lock to prevent automatic updates from interfering with other dialogs
        if self.UpdateLock:
            return True
        self.LockUpdates()
        for Project in self.Data.Projects:
            for Experiment in Project.Children:
                for Run in Experiment.Children:
                    for Job in Run.Children:
                        self.Data.CurrentJob = Job
                        self.GetJobStatus()
        self.UnlockUpdates()
        return True
    
    def GetJobStatus (self, widget=None, data=None, ArchivePrompt=True):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        if (self.Data.CurrentJob.RunStatus == SM_STATUS_NOTAVAIL):
            self.NotifyDialog("Job has not been submitted yet.")
            return True
        if (self.Data.CurrentJob.RunStatus == SM_STATUS_DONE):
#            self.GetJobLog(self.Data.CurrentJob)
#            self.GetJobOutput(self.Data.CurrentJob)
            return True
        OldStatus = self.Data.CurrentJob.RunStatus
        machine = self.Data.Machines.EntryByName(self.Data.CurrentJob.ExecMachine)
        system  = self.Data.QueuingSystemByName(machine.QueueSystem)
        command = system.GetStatusCommand(self.Data.CurrentJob)
        result, output = self.RemoteExec(machine, "", command)
        NewStatus = system.GetJobStatus(output)
        if ((NewStatus == SM_STATUS_UNK) or (result == SM_FAIL)):
            override_result = self.QuestionDialog('Unable to determine the status of job "'+ \
                                             self.Data.CurrentJob.Name+'" (remote ID '+ \
                                             self.Data.CurrentJob.RemoteJobID+'). '+ \
                                             'This can happen because of a problem with '+\
                                             'the remote job scheduler, in which case you'+\
                                             ' should try again later. It may also simply mean'+\
                                             ' that the job has completed and the queuing system'+\
                                             ' has forgotten about it (e.g., with PBS). Would you'+\
                                             ' like to mark this job as complete?')
            if (override_result == gtk.RESPONSE_OK):
                result = SM_OK
                NewStatus = SM_STATUS_DONE
        if (result == SM_OK):
            self.Data.CurrentJob.RunStatus = NewStatus
            self.UpdateTreeStoreRow(self.Data.CurrentJob)
            if isinstance(widget, gtk.Button):
                self.SetDialogData(self.JobDialog, self.Data.CurrentJob)
            if ArchivePrompt and \
               ((OldStatus != self.Data.CurrentJob.RunStatus) and \
                (self.Data.CurrentJob.RunStatus == SM_STATUS_DONE)):
                    self.GetJobLog(self.Data.CurrentJob)
                    self.GetJobOutput(self.Data.CurrentJob)
                    DataDestMachine = self.Data.Machines.EntryByName(self.Data.CurrentJob.DataDestMachine)
                    if DataDestMachine:
                        result = self.QuestionDialog('Job "'+self.Data.CurrentJob.Name+'" (remote ID '+ \
                                                     self.Data.CurrentJob.RemoteJobID+') has finished.'+ \
                                                     ' Perform post-run data transfers now?')
                        if (result == gtk.RESPONSE_OK):
                            self.ArchiveJobData(self.Data.CurrentJob)
        else:
            self.NotifyDialog('Job status check for job "'+self.Data.CurrentJob.Name+\
            			'" (remote ID '+self.Data.CurrentJob.RemoteJobID+\
                                ') failed. Previous value is unchanged.')
        return True

    def ContinueJob (self, widget, data=None):
        if isinstance(widget, gtk.Button):
            self.GetDialogData(self.JobDialog, self.Data.CurrentJob)
        if ((self.Data.CurrentJob.RunStatus != SM_STATUS_DONE) and \
            (self.Data.CurrentJob.RunStatus != SM_STATUS_UNK)):
            self.NotifyDialog("Cannot continue a job that has not yet finished.")
            return True
        if (self.Data.CurrentJob.RunStatus == SM_STATUS_UNK):
            result = self.QuestionDialog("Job status is unknown. Are you sure you want to create a restart job?")
            if (result == gtk.RESPONSE_CANCEL):
                return True
        try:
            app = self.Data.Applications.EntryByName(self.Data.CurrentJob.Application)
            if not app:
                self.NotifyDialog("No application defined for this job!  How did that happen?")
                return True
            if (app.RestartType == SM_RESTART_COPY):
                machine = self.Data.Machines.EntryByName(self.Data.CurrentJob.ExecMachine)
                if not machine:
                    self.NotifyDialog("No execution machine defined for this job!  How did that happen?")
                    return True
                result = self.TransferFiles(machine, \
                                            [self.Data.RemotePath([self.Data.CurrentJob.RemoteRunDir,\
                                                                   app.RestartCopyFromFile], machine.Name)], \
                                            SM_LOCALHOST, TempFile, True, self.wTree.window)
                if (result == SM_OK):
                    f = open(TempFile)
                    temp_data = f.read()
                    f.close()
                else:
                    return True
            else:
                temp_data = ""
            job = self.Data.CurrentJob.CreateRestartJob(self.Data, temp_data)
            self.AddTreeStoreRow(self.Data.CurrentJob.Parent, job.Name, job)
            self.Data.Changed = True
        except IOError:
            self.NotifyDialog("Error writing local temp file.")
        return True

    def EditContainerDialog (self, Dialog, Container, Choices=None):
        self.SetDialogData(Dialog, Container, Choices)
        result = None
        while ((result != gtk.RESPONSE_CANCEL) & (result != gtk.RESPONSE_OK) &
               (result != gtk.RESPONSE_DELETE_EVENT)):
            Backup = Container.Backup()
            result = Dialog.dialog.run()
            if (result != gtk.RESPONSE_CANCEL):
                self.GetDialogData(Dialog, Container)
            else:
                Container.RestoreBackup(Backup)
            self.UpdateAllTreeStoreRows()
#        self.Data.PrintAll()
        Dialog.dialog.hide()
        return result

    def EditProjectData (self, widget=None, data=None):
        result = self.EditContainerDialog(self.ProjectDialog, self.Data.CurrentProject)
        if widget:
            return True
        else:
            return result

    def EditExperimentData (self, widget=None, data=None):
        AppNames  = self.Data.Applications.EntryNames()
        MachNames = self.Data.Machines.EntryNames()
        ExecMachine     = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.ExecMachine)
        DataSrcMachine  = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.DataSrcMachine)
        DataDestMachine = self.Data.Machines.EntryByName(self.Data.CurrentExperiment.DataDestMachine)
        if ExecMachine:
            Queues   = ExecMachine.Queues
            Accounts = ExecMachine.Accounts
        else:
            Queues   = [ ]
            Accounts = [ ]
        if (self.Data.CurrentExperiment.ConfigStatus == SM_OK):
            result = self.ExperimentDialog.ConfigButton.modify_bg(gtk.STATE_NORMAL, \
                                                                  gtk.gdk.color_parse("green"))
        elif (self.Data.CurrentExperiment.ConfigStatus == SM_FAIL):
            result = self.ExperimentDialog.ConfigButton.modify_bg(gtk.STATE_NORMAL, \
                                                                  gtk.gdk.color_parse("red"))
        if (self.Data.CurrentExperiment.BuildStatus == SM_OK):
            result = self.ExperimentDialog.BuildButton.modify_bg(gtk.STATE_NORMAL, \
                                                                 gtk.gdk.color_parse("green"))
        elif (self.Data.CurrentExperiment.BuildStatus == SM_FAIL):
            result = self.ExperimentDialog.BuildButton.modify_bg(gtk.STATE_NORMAL, \
                                                                 gtk.gdk.color_parse("red"))
        result = self.EditContainerDialog(self.ExperimentDialog, self.Data.CurrentExperiment, \
                                 {"Application"     : AppNames, \
                                  "ExecMachine"     : MachNames, \
                                  "ExecQueue"       : Queues, \
                                  "ExecAccount"     : Accounts, \
                                  "DataSrcMachine"  : MachNames, \
                                  "DataDestMachine" : MachNames})
        self.ExperimentDialog.ConfigButton.modify_style(self.ExperimentDialog.SavedButtonStyle)
        self.ExperimentDialog.BuildButton.modify_style(self.ExperimentDialog.SavedButtonStyle)
        if widget:
            return True
        else:
            return result
                                
    def EditRunData (self, widget=None, data=None):
        AppNames  = self.Data.Applications.EntryNames()
        MachNames = self.Data.Machines.EntryNames()
        ExecMachine     = self.Data.Machines.EntryByName(self.Data.CurrentRun.ExecMachine)
        DataSrcMachine  = self.Data.Machines.EntryByName(self.Data.CurrentRun.DataSrcMachine)
        DataDestMachine = self.Data.Machines.EntryByName(self.Data.CurrentRun.DataDestMachine)
        if ExecMachine:
            Queues   = ExecMachine.Queues
            Accounts = ExecMachine.Accounts
        else:
            Queues   = [ ]
            Accounts = [ ]
        result = self.EditContainerDialog(self.RunDialog, self.Data.CurrentRun, \
                                 {"Queue"           : Queues, \
                                  "Account"         : Accounts, \
                                  "DataSrcMachine"  : MachNames, \
                                  "DataDestMachine" : MachNames})
        if widget:
            return True
        else:
            return result

    def EditJobData (self, widget=None, data=None):
        AppNames  = self.Data.Applications.EntryNames()
        MachNames = self.Data.Machines.EntryNames()
        ExecMachine     = self.Data.Machines.EntryByName(self.Data.CurrentJob.ExecMachine)
        DataSrcMachine  = self.Data.Machines.EntryByName(self.Data.CurrentJob.DataSrcMachine)
        DataDestMachine = self.Data.Machines.EntryByName(self.Data.CurrentJob.DataDestMachine)
        if ExecMachine:
            Queues   = ExecMachine.Queues
            Accounts = ExecMachine.Accounts
        else:
            Queues   = [ ]
            Accounts = [ ]
        result = self.EditContainerDialog(self.JobDialog, self.Data.CurrentJob, \
                                 {"Queue"           : Queues, \
                                  "Account"         : Accounts, \
                                  "DataSrcMachine"  : MachNames, \
                                  "DataDestMachine" : MachNames})
        if widget:
            return True
        else:
            return result

    def DeleteContainerDialog (self, container, typename, confirm=True):
        if (confirm and self.Data.CurrentUser.ConfirmDeletions):
            dTree  = gtk.glade.XML(GladeFile, "DeleteConfirmDialog")
            dialog = dTree.get_widget("DeleteConfirmDialog")
            dialog.hide()
            label  = dTree.get_widget("label189")
            label.set_text(typename + ' "' + container.Name + '"')
            dialog.show()
            result = dialog.run()
            dialog.destroy()
        else:
            result = gtk.RESPONSE_OK
        if (result == gtk.RESPONSE_OK):
            treeiter = self.FindTreeIter(self.treestore.iter_children(None), \
                                         container)
            self.RemoveTreeStoreRow(treeiter)
            if (typename == "Project"):
                self.Data.RemoveProjectFromList(container)
            container.Delete()
            self.Data.Changed = True

    def DeleteProject (self, widget=None, data=None, confirm=True):
        self.DeleteContainerDialog(self.Data.CurrentProject, "Project", confirm)

    def DeleteExperiment (self, widget=None, data=None, confirm=True):
        self.DeleteContainerDialog(self.Data.CurrentExperiment, "Experiment", confirm)

    def DeleteRun (self, widget=None, data=None, confirm=True):
        self.DeleteContainerDialog(self.Data.CurrentRun, "Run", confirm)

    def DeleteJob (self, widget=None, data=None, confirm=True):
        self.DeleteContainerDialog(self.Data.CurrentJob, "Job", confirm)

    def CloneProject (self, widget=None, data=None):
        self.Data.CloneProject(self.Data.CurrentProject)
        self.AddTreeStoreRow(None, self.Data.CurrentProject.Name, \
                             self.Data.CurrentProject)
        self.Data.Changed = True

    def CloneExperiment (self, widget=None, data=None):
        self.Data.CloneExperiment(self.Data.CurrentExperiment)
        self.AddTreeStoreRow(self.Data.CurrentExperiment.Parent, \
                             self.Data.CurrentExperiment.Name, \
                             self.Data.CurrentExperiment)
        self.Data.Changed = True

    def CloneRun (self, widget=None, data=None):
        self.Data.CloneRun(self.Data.CurrentRun)
        self.AddTreeStoreRow(self.Data.CurrentRun.Parent, \
                             self.Data.CurrentRun.Name, \
                             self.Data.CurrentRun)
        self.Data.Changed = True

    def CloneJob (self, widget=None, data=None):
        self.Data.CloneJob(self.Data.CurrentJob)
        self.AddTreeStoreRow(self.Data.CurrentJob.Parent, \
                             self.Data.CurrentJob.Name, \
                             self.Data.CurrentJob)
        self.Data.Changed = True

    def FileChooserDialog (self, filename, dialogname, title="Browse"):
        self.cTree   = gtk.glade.XML(GladeFile, dialogname)
        self.chooser = self.cTree.get_widget(dialogname)
        self.chooser.set_filename(filename)
        self.chooser.set_title(title)
        result = self.chooser.run()
        if (result == gtk.RESPONSE_OK):
            filename = self.chooser.get_filename()
        self.chooser.destroy()
        return result, filename

    def OpenProjectFile (self, widget, data=None):
        if self.Data.Changed:
            result = self.CloseProjectFile()
            if (result == SM_FAIL):
                return True
        result, filename = self.FileChooserDialog(self.Data.CurrentFile, \
                                                  "FileChooserDialog", "Open project file")
        if (result == gtk.RESPONSE_OK):
            self.Data.Clear()
            self.Data.CurrentFile = filename
            result = self.Data.ReadFromDisk(self.Data.CurrentFile)
            self.MakeTreeViewConsistent()
            if (result == SM_FAIL):
                self.NotifyDialog("Failed to read file.")
            else:
                self.treestore.clear()
                for Project in self.Data.Projects:
                    self.AddTreeStoreRow(None, Project.Name, Project)
                self.ApplyUserSettings()
        return True

    def MergeProjectFile (self, widget, data=None):
        mergefile = self.Data.CurrentUser.Workspace
        result, filename = self.FileChooserDialog(mergefile, "FileChooserDialog", "Merge project file")
        if (result == gtk.RESPONSE_OK):
            mergefile = filename
            result = self.Data.ReadFromDisk(mergefile, merge=True)
            self.MakeTreeViewConsistent()
            if (result == SM_FAIL):
                self.NotifyDialog("Failed to read file.")
            else:
                self.treestore.clear()
                for Project in self.Data.Projects:
                    self.AddTreeStoreRow(None, Project.Name, Project)
                self.Data.Changed = True
        return True

    def SaveProjectFile (self, widget=None, data=None):
        if not self.Data.FileNameSet:
            result = self.SaveProjectFileAsNew(widget, data)
        else:
            result = self.Data.SaveToDisk()
            if (result == SM_FAIL):
                self.NotifyDialog("Failed to write file.")
        if widget:   # we were called as an event handler
            return True
        else:
            return result

    def SaveProjectFileAsNew (self, widget=None, data=None):
        result, filename = self.FileChooserDialog(self.Data.CurrentFile, \
                                                  "SaveFileChooserDialog")
        if (result == gtk.RESPONSE_OK):
            if os.path.isfile(filename):
                result = self.QuestionDialog("A file with this name already exists. Overwrite?")
            if (result == gtk.RESPONSE_OK):
                self.Data.CurrentFile = filename
                self.Data.FileNameSet = True
                result = self.Data.SaveToDisk()
                if (result == SM_FAIL):
                    self.Data.FileNameSet = False
                    self.NotifyDialog("Failed to write file.")
            else:
                result = SM_FAIL
        else:
            result = SM_FAIL
        if widget:   # we were called as an event handler
            return True
        else:
            return result

    def AboutDialog (self, widget, data=None):
        self.dTree  = gtk.glade.XML(GladeFile, "AboutDialog")
        self.dialog = self.dTree.get_widget("AboutDialog")
        result = self.dialog.run()
        self.dialog.destroy()
    
    def TreeViewButtonPress (self, widget, event):
        if (event.button == 3):
            treeselection = self.treeview.get_selection()
            treepath = self.treeview.get_path_at_pos(int(event.x), \
                                                     int(event.y))
            if (treepath != None):
                treepath = treepath[0]
                treeselection.unselect_all()
                treeselection.select_path(treepath)
                self.InvokeOptionsMenu(widget, event)
            return True
        else:
            return False

    def TreeViewDoubleClick (self, treeview, treepath, view_column):
        container = self.TreeViewSelection()
        if isinstance(container, Project):
            self.Data.CurrentProject = container
            self.EditProjectData()
        elif isinstance(container, Experiment):
            self.Data.CurrentExperiment = container
            self.EditExperimentData()
        elif isinstance(container, Run):
            self.Data.CurrentRun = container
            self.EditRunData()
        elif isinstance(container, Job):
            self.Data.CurrentJob = container
            self.EditJobData()
        return True
        
    def TreeViewSelection (self):
        treeselection = self.treeview.get_selection()
        (treemodel, treeiter) = treeselection.get_selected()
        object = treemodel.get_value(treeiter, 1)
        return object

    def InvokeOptionsMenu (self, widget, event=None):
        container = self.TreeViewSelection()
        if container:
            self.OptionsMenu(container, event)
            return True
        else:
            return False

    # need to properly objectify this to handle Experiments, Runs, Jobs

    def OptionsMenu (self, container, event):
        if isinstance(container, Project):
            self.mTree = gtk.glade.XML(GladeFile, "ProjectOptionsMenu")
            self.Data.CurrentProject = container
            self.mTree.signal_autoconnect( \
                { "on_edit1_activate"           : self.EditProjectData, \
                  "on_new_experiment1_activate" : self.AddExperiment, \
                  "on_take_notes1_activate"     : self.InvokeNoteTaker, \
                  "on_clone1_activate"          : self.CloneProject, \
                  "on_delete2_activate"         : self.DeleteProject, \
                } )
            self.menu = self.mTree.get_widget("ProjectOptionsMenu")
        elif isinstance(container, Experiment):
            self.mTree = gtk.glade.XML(GladeFile, "OptionsMenu")
            self.Data.CurrentExperiment = container
            self.mTree.signal_autoconnect( \
                { "on_edit1_activate"   : self.EditExperimentData, \
                  "on_clone1_activate"  : self.CloneExperiment, \
                  "on_delete2_activate" : self.DeleteExperiment, \
                } )
            self.menu = self.mTree.get_widget("OptionsMenu")
        elif isinstance(container, Run):
            self.mTree = gtk.glade.XML(GladeFile, "RunOptionsMenu")
            self.Data.CurrentRun = container
            self.mTree.signal_autoconnect( \
                { "on_edit1_activate"            : self.EditRunData, \
                  "on_new_job1_activate"         : self.AddJob, \
                  "on_view_parameters1_activate" : self.ViewRunParameters, \
                  "on_clone1_activate"           : self.CloneRun, \
                  "on_delete2_activate"          : self.DeleteRun, \
                } )
            self.menu = self.mTree.get_widget("RunOptionsMenu")
        elif isinstance(container, Job):
            self.mTree = gtk.glade.XML(GladeFile, "JobOptionsMenu")
            self.Data.CurrentJob = container
            self.mTree.signal_autoconnect( \
                { "on_edit1_activate"            : self.EditJobData, \
                  "on_submit_job1_activate"      : self.SubmitJob, \
                  "on_update_status1_activate"   : self.GetJobStatus, \
                  "on_archive_data1_activate"    : self.ArchiveJobDataFromButton, \
                  "on_view_parameters1_activate" : self.ViewJobParameters, \
                  "on_view_log1_activate"        : self.ViewJobLog, \
                  "on_view_output1_activate"     : self.ViewJobOutput, \
                  "on_continue_job1_activate"    : self.ContinueJob, \
                  "on_clone1_activate"           : self.CloneJob, \
                  "on_delete2_activate"          : self.DeleteJob, \
                } )
            self.menu = self.mTree.get_widget("JobOptionsMenu")
        if (event != None):
            self.menu.popup(None, None, None, event.button, event.time)
        else:
            self.menu.popup(None, None, None, 3, 0)
        return True

    def EditTextFile (self, file):
        caps = mailcap.getcaps()
        cmd, mc = mailcap.findmatch(caps, "text/plain", filename=file)
        if (cmd):
            os.system(cmd + " &")
        else:
            self.NotifyDialog("No default text editor defined! Check your operating system's " + \
                              "application settings for MIME type text/plain.")

    def ContainerDialogObject (self, container):
        if isinstance(container, Project):
            return self.ProjectDialog
        elif isinstance(container, Experiment):
            return self.ExperimentDialog
        elif isinstance(container, Run):
            return self.RunDialog
        elif isinstance(container, Job):
            return self.JobDialog
        else:
            return None
    
    def InvokeNoteTaker (self, widget, event=None):
        container = self.TreeViewSelection()
        if container:
            dialog = self.ContainerDialogObject(container)
            self.GetDialogData(dialog, container)
            self.TakeNotes(container)
            self.SetDialogData(dialog, container)
            return True
        else:
            return False

    def TakeNotes (self, container):
        if (container.NotesType == SM_WIKI):
            self.Data.LogMessage("Launch web browser with URL: %s" % container.NotesLocation)
            webbrowser.open(container.NotesLocation)
        else:
            self.Data.LogMessage("Launch text editor with file: %s" % container.NotesLocation)
            self.EditTextFile(container.NotesLocation)

    def DeleteEvent (self, widget, event, data=None):
        if self.Data.Changed:
            result = self.SaveQuestionDialog("Some data unsaved. Really quit?")
        else:
            result = self.QuitDialog()
        if (result == gtk.RESPONSE_NO):
            save_result = self.SaveProjectFile()
        else:
            save_result = SM_OK
        if ((result == gtk.RESPONSE_OK) or (result == gtk.RESPONSE_NO and save_result == SM_OK)):
            self.Data.Finalize()
            gtk.main_quit()
            return False
        else:
            return True

    def CloseProjectFile (self, widget=None, data=None):
        if self.Data.Changed:
            result = self.SaveQuestionDialog("Some data unsaved. Really close current project file?", \
                                             dialog="SaveCloseQuestionDialog", label="label297")
        else:
            result = gtk.RESPONSE_OK
        if (result == gtk.RESPONSE_NO):
            save_result = self.SaveProjectFile()
        else:
            save_result = SM_OK
        if ((result == gtk.RESPONSE_OK) or (result == gtk.RESPONSE_NO and save_result == SM_OK)):
            ProjectList = copy.copy(self.Data.Projects)
            for Project in ProjectList:
                self.DeleteContainerDialog(Project, "Project", confirm=False)
            self.Data.Clear()
            self.MakeTreeViewConsistent()
            result = SM_OK
        else:
            result = SM_FAIL
        if widget:    # called as a signal handler
            return True
        else:         # called as a regular method
            return result

    def Quit (self, widget=None, data=None):
        if self.Data.Changed:
            result = self.SaveQuestionDialog("Some data unsaved. Really quit?")
        else:
            result = self.QuitDialog()
        if (result == gtk.RESPONSE_NO):
            save_result = self.SaveProjectFile()
        else:
            save_result = SM_OK
        if ((result == gtk.RESPONSE_OK) or (result == gtk.RESPONSE_NO and save_result == SM_OK)):
            self.Data.Finalize()
            gtk.main_quit()

# Data store class -- holds all current resource and container data independent
# of the GUI

class AppStore:
    def __init__ (self):
        self.Projects          = [ ]
        self.FileNameSet       = False
        self.CurrentUser       = User()
        self.CurrentFile       = os.path.join(self.CurrentUser.Workspace, "Default.teu")
        self.CurrentProject    = Project(self.CurrentUser)
        self.CurrentExperiment = Experiment(self.CurrentUser, self.CurrentProject)
        self.CurrentRun        = Run(self.CurrentUser, self.CurrentExperiment)
        self.CurrentJob        = Job(self.CurrentUser, self.CurrentRun, "New Job")
        self.LocalMethod       = AccessMethod("local", "", '/bin/sh -c "%c"', 'scp -B -r -p -q "%f" "%r"', \
                                              'scp -B -r -p -q "%r" "%f"', SM_AUTH_HOST, "", "", False)
        self.SSHMethod         = AccessMethod("ssh", "", "ssh %u@%h '%c'", \
        				      'scp -r -p -q "%f" %u@%h:"%r"', \
                                              'scp -r -p -q %u@%h:"%r" "%f"', SM_AUTH_PASSWORD, \
                                              "Password", "assword:", False)
        self.SSHRSAMethod      = AccessMethod("ssh-rsa", "", "ssh -o BatchMode=yes %u@%h '%c'", \
        				      'scp -B -r -p -q "%f" %u@%h:"%r"', \
                                              'scp -B -r -p -q %u@%h:"%r" "%f"', SM_AUTH_HOST, "", "", False)
        self.SSHAgentMethod    = AccessMethod("ssh-agent", "ssh-add", "ssh -a -o BatchMode=yes %u@%h '%c'", \
        				      'scp -B -r -p -q "%f" %u@%h:"%r"', \
                                              'scp -B -r -p -q %u@%h:"%r" "%f"', SM_AUTH_KEYAGENT, "Passphrase", \
                                              r'Enter passphrase for .+:', False)
        self.GSISSHMethod      = AccessMethod("gsissh", "grid-proxy-init -cert %K -key %k -valid %l:00", \
                                              "gsissh -o BatchMode=yes -o GSSAPIAuthentication=yes %u@%h '%c'", \
        				      'gsiscp -B -r -p -q "%f" %u@%h:"%r"', \
                                              'gsiscp -B -r -p -q %u@%h:"%r" "%f"', \
                                              SM_AUTH_GSI, "Passphrase", \
                                              "Enter GRID pass phrase for this identity:", False, \
                                              'gsiscp -B -r -p -q %u@%h:"%f" %v@%i:"%r"')
        self.SSHGlobusMethod   = AccessMethod("gsissh+globus", "grid-proxy-init -cert %K -key %k -valid %l:00", \
                                              "gsissh -o BatchMode=yes -o GSSAPIAuthentication=yes %u@%h '%c'", \
        				      'globus-url-copy -p 1 -tcp-bs 8388608 file:"%f" gsiftp://%h/"%r"', \
                                              'globus-url-copy -p 1 -tcp-bs 8388608 gsiftp://%h/"%r" file:"%f"', \
                                              SM_AUTH_GSI, "Passphrase", \
                                              "Enter GRID pass phrase for this identity:", False, \
                                              'globus-url-copy -p 1 -tcp-bs 8388608 gsiftp://%h/"%f" gsiftp://%i/"%r"')
        self.SSHUberFTPMethod  = AccessMethod("gsissh+uberftp", "grid-proxy-init -cert %K -key %k -valid %l:00", \
                                              "gsissh -o BatchMode=yes -o GSSAPIAuthentication=yes %u@%h '%c'", \
        				      'gsiscp -B -r -p -q "%f" %u@%h:"%r"', \
                                              'gsiscp -B -r -p -q %u@%h:"%r" "%f"', \
#        				      'uberftp %h \"put %f %r\"', \
#                                             'uberftp %h \"get %r %f\"', \
                                              SM_AUTH_GSI, "Passphrase", \
                                              "Enter GRID pass phrase for this identity:", False, \
                                              'uberftp %i \"lopen %h; put %f %r\"')
        self.KerberosMethod    = AccessMethod("kerberos", "kinit %u@%R", "rsh %h -l %u '%c'", \
                                              'rcp -r -p "%f" %u@%h:"%r"', \
                                              'rcp -r -p %u@%h:"%r" "%f"', SM_AUTH_KERBEROS, \
                                              "Password", "Password", False)
        self.AccessMethods     = [self.LocalMethod, self.SSHMethod, self.SSHRSAMethod, \
                                  self.SSHAgentMethod, self.KerberosMethod, self.GSISSHMethod, \
                                  self.SSHGlobusMethod, self.SSHUberFTPMethod]
        self.AccessNames       = [ ]
        for Method in self.AccessMethods:
            self.AccessNames.append(Method.Name)
        self.UnixNoQueue       = QueuingSystem("unix", "python %b", "ps -p %j -o stat=", "kill %j", "", \
  					       "#!/usr/bin/env python\nimport os, os.path, sys\n" + \
                                               "if (not os.path.exists('%d')): os.makedirs('%d')\n" + \
                                               "os.chdir('%d')\n" + \
                                               "pid = os.spawnvp(os.P_NOWAIT, '/bin/sh', ['/bin/sh', '-c', '%P %e %a 1>%z/%L.out 2>%z/%L.err'])\n" + \
                                               "print 'pid ', pid\nsys.exit(0)", \
                                               r'pid  ([0-9]+)', "%z/%L.out", "%z/%L.err", r'^.+', \
                                               {"R": SM_STATUS_RUN, "S": SM_STATUS_PEND})
        self.PBSQueue          = QueuingSystem("pbs", "qsub %b", "qstat %j", "qdel %j", "", \
                                               "#!/bin/sh\n#PBS -l walltime=%H:%M:%S\n" + \
                                               "#PBS -l mem=%mmb\n#PBS -l ncpus=%n\n" + \
                                               "#PBS -q %Q\n#PBS -V\n#PBS -A %A\n#PBS -N %L\n" + \
                                               "%D %d\n%C %d\n%P %e %a", \
                                               r'(?m)^([0-9]+).*\b', "%z/%L.o%j", "%z/%L.e%j", \
                                               r'(?m)^[0-9]+.*\b\b.+\b\b.+\b\b.+\b\b([A-Z])\b.*', \
                                               {"Q": SM_STATUS_PEND, "R": SM_STATUS_RUN}
                                               )
        self.LSFQueue          = QueuingSystem("lsf", "bsub < %b", "bjobs %j", "bkill %j", "bpeek %j", \
                                               "#!/bin/sh\n#BSUB -W %t\n#BSUB -q %Q\n#BSUB -n %n\n" + \
                                               "#BSUB -J %L\n#BSUB -P %A\n#BSUB -e %z/%J.err\n#BSUB -o %z/%J.out\n" + \
                                               "%D %d\n%C %d\n%P %e %a", \
                                               r'Job <(.+)> is submitted to queue', \
                                               "%z/%j.out", "%z/%j.err", \
                                               r'.+\b\b.+\b\b([A-Z]+)\b\b.+\b\b.+\b\b.+\b\b.+\b\b[0-9]+\b.*', \
                                               {"PEND": SM_STATUS_PEND, "RUN": SM_STATUS_RUN, \
                                                "DONE": SM_STATUS_DONE})
        self.LoadLevelerQueue  = QueuingSystem("loadleveler", "llsubmit %b", "llq %j", "llcancel -q %j", "", \
                                               "#!/bin/sh\n#@ wall_clock_limit = %t\n" + \
                                               "#@ environment = COPY_ALL\n#@ job_type = parallel\n" + \
                                               "#@ class = batch\n#@ account_no = %A\n#@ tasks_per_node = %T\n" + \
                                               "#@ job_name = %L\n#@ output = %z/$(job_name).$(jobid).out\n" + \
                                               "#@ error = %z/$(job_name).$(jobid).err\n#@ queue\n" + \
                                               "%D %d\n%C %d\n%P %e %a", \
                                               r'llsubmit: The job "(.+)" has been submitted', \
                                               "%z/%L.%j.out", "%z/%L.%j.err", \
                                               r'.+\b\b.+\b\b.+\b\b.+\b\b([A-Z])\b.*', \
                                               {"P": SM_STATUS_PEND, "R": SM_STATUS_RUN, \
                                                "C": SM_STATUS_DONE})
        self.QueuingSystems    = [self.UnixNoQueue, self.PBSQueue, self.LSFQueue, self.LoadLevelerQueue]
        self.QueueNames        = [ ]
        for Queue in self.QueuingSystems:
            self.QueueNames.append(Queue.Name)
        self.Applications      = ApplicationList(self)
        self.Machines          = MachineList(self)
        self.Changed           = False
    def PrintAll (self):
        print "Container tree listing:"
        for Project in self.Projects:
            Project.Print()
    def AddProject (self):
        self.CurrentProject = Project(self.CurrentUser)
        # if we didn't press cancel...
        self.Projects.append(self.CurrentProject)
    def RemoveProjectFromList (self, project):
        self.Projects.remove(project)
    def CloneProject (self, project):
        clone = project.Clone(self.CurrentUser)
        self.Projects.append(clone)
        self.CurrentProject = clone
    def CloneExperiment (self, experiment):
        clone = experiment.Clone(self.CurrentUser)
        self.CurrentExperiment = clone
    def CloneRun (self, run):
        clone = run.Clone(self.CurrentUser)
        self.CurrentRun = clone
    def CloneJob (self, job):
        clone = job.Clone(self.CurrentUser)
        self.CurrentJob = clone
    def AddExperiment (self):
        self.CurrentExperiment = self.CurrentProject.AddChild(self.CurrentUser)
    def Clear (self):
        for Project in self.Projects:
            Project.Delete()
        self.__init__()
    def LogMessage (self, text):
        if self.CurrentUser.VerboseLogging:
            logfile = os.path.join(self.CurrentUser.Workspace, self.CurrentUser.LogFile)
            try:
                if os.path.exists(logfile):
                    f = open(logfile, 'a')
                else:
                    f = open(logfile, 'w')
                f.write(text + "\n")
                f.close()
            except IOError:
                pass
    def ReadFromDisk (self, filename, merge=False):
        self.LogMessage("Read AppStore data from file %s" % filename)
        try:
            f = open(filename)
            store = AppStore()
            store.__dict__ = cPickle.load(f)
            f.close()
            if merge:
                self.Projects.extend(store.Projects)
                self.Applications.Extend(store.Applications)
                self.Machines.Extend(store.Machines)
#               self.AccessMethods.extend(store.AccessMethods)
#               self.AccessNames = [ ]
#               for Method in self.AccessMethods:
#                   self.AccessNames.append(Method.Name)
#               self.QueuingSystems.extend(store.QueuingSystems)
#               self.QueueNames = [ ]
#               for Queue in self.QueuingSystems:
#                   self.QueueNames.append(Method.Name)
            else:
                self.Projects     = store.Projects
                self.Applications = store.Applications
                self.Machines     = store.Machines
                self.CurrentUser  = store.CurrentUser
            self.CurrentFile = filename
            self.FileNameSet = True
            self.Changed     = False
            return SM_OK
        except IOError:
            return SM_FAIL
        except cPickle.UnpicklingError:
            return SM_FAIL
    def SaveToDisk (self):
        self.LogMessage("Save AppStore data to file %s" % self.CurrentFile)
        try:
            f = open(self.CurrentFile, "w")
            # Make sure we do not store any valid AuthData in the file,
            # but don't throw any out of memory when we save
            MachCopy = copy.deepcopy(self.Machines)
            for Machine in self.Machines.Entries:
                Machine.AuthData        = ""
                Machine.AlreadyPrompted = False
                Method = self.AccessMethodByName(Machine.AccessMethod)
                print "Clearing: ", Machine.Name, Machine.AccessMethod
                Machine.StoreAuthData   = Method.StoreAuthData
            cPickle.dump(self.__dict__, f)
            f.close()
            del self.Machines
            self.Machines = MachCopy
            self.Changed  = False
            return SM_OK
        except IOError:
            return SM_FAIL
    def AccessMethodByName (self, name):
        result = None
        for method in self.AccessMethods:
            if (method.Name == name):
                result = method
        return result
    def QueuingSystemByName (self, name):
        result = None
        for system in self.QueuingSystems:
            if (system.Name == name):
                result = system
        return result
    def CancelRemoteExec (self):
        if self.LiveRemoteProcess:
            self.LogMessage("Cancelling remote process.")
            self.LiveRemoteProcess.kill(9)
    def EscapeProblemChars (self, string):
        output = string
        output = output.replace("$", "\\$")
        output = output.replace(" ", "\\ ")
        output = output.replace("\r", "")
        output = output.replace("\n", "")
        return output
    def TransferFile (self, SourceMachine, SourceFile, DestMachine, DestPath):
        if (SourceMachine == DestMachine) or \
           ((SourceMachine.AccessMethod == "local") and (DestMachine.AccessMethod == "local")):
                                                                  # single-party data transfer
            if (SourceMachine.AccessMethod == "local"):
                try:
                    self.LogMessage("Copying %s to %s" % (SourceFile, DestPath))
                    shutil.copy2(SourceFile, DestPath)
                    result, output = (SM_OK, "")
                except IOError:
                    result, output = (SM_FAIL, "TransferFile:  I/O error on local copy")
            else:
                src            = self.EscapeProblemChars(SourceFile)
                dest           = self.EscapeProblemChars(DestPath)
                command        = SourceMachine.GetFileCopyCommand(src, dest)
                result, output = self.RemoteExec(SourceMachine.Name, None, command)
        elif (SourceMachine.AccessMethod == "local"):             # second-party, upload
            method         = self.AccessMethodByName(DestMachine.AccessMethod)
            dest           = self.EscapeProblemChars(DestPath)
            result, output = method.UploadFile(DestMachine.RemoteUserid, DestMachine.LoginHost, \
                                               DestMachine.AuthData, SourceFile, dest, \
                                               DestMachine, self)
        elif (DestMachine.AccessMethod == "local"):               # second-party, download
            method         = self.AccessMethodByName(SourceMachine.AccessMethod)
            src            = self.EscapeProblemChars(SourceFile)
            result, output = method.DownloadFile(SourceMachine.RemoteUserid, SourceMachine.LoginHost, \
                                                 SourceMachine.AuthData, DestPath, src, \
                                                 SourceMachine, self)
        else:                                                     # third-party data transfer
            if (SourceMachine.AccessMethod != DestMachine.AccessMethod):
                result, output = (SM_FAIL, "TransferFile:  source and dest access methods not the same")
            else:
                method         = self.AccessMethodByName(SourceMachine.AccessMethod)
                src            = self.EscapeProblemChars(SourceFile)
                dest           = self.EscapeProblemChars(DestPath)
                result, output = method.TransferThirdPartyFile(SourceMachine.RemoteUserid, DestMachine.RemoteUserid, \
                                                               SourceMachine.AuthData, \
                                                               SourceMachine.LoginHost, DestMachine.LoginHost, \
                                                               src, dest, SourceMachine, DestMachine, self)
        if not SourceMachine.StoreAuthData:
            SourceMachine.AuthData = ""
        if not DestMachine.StoreAuthData:
            DestMachine.AuthData = ""
        return (result, output)
    def LocateGridIdentityFiles (self, subject, path):
        if not os.path.isdir(path):
            return None, None
        entries = os.listdir(path)
        for entry in entries:
            ext  = os.path.splitext(entry)[1].lower()
            full = os.path.join(path, entry)
            if os.path.isfile(full) and (ext == ".pem"):
                try:
                    process = pexpect.spawn("grid-cert-info -subject -file "+full)
                    process.expect(pexpect.EOF)
                    output = process.before
                    process.close(wait=1)
                    if (output.count(subject) > 0):
                        certfile = full
                        keyfile  = certfile.replace("cert.pem", "key.pem")
                        if not os.path.isfile(keyfile):  keyfile = None
                        return certfile, keyfile
                except pexpect.ExceptionPexpect:
                    break
        return None, None
    def MakeRemoteDirectory (self, machine, path):
        if (machine.AccessMethod == "local"):
            if os.path.exists(path):
                return (SM_OK, "")
            try:
                os.makedirs(path)
                return (SM_OK, "")
            except OSError:
                return (SM_FAIL, "MakeRemoteDirectory:  OS error on local makedir")
        else:
            command        = machine.GetMakeDirCommand(path)
            result, output = self.RemoteExec(machine.Name, "", command)
            return (result, output)
    def RemoteExec (self, machname, startdir, command, store_auth=False):
        machine  = self.Machines.EntryByName(machname)
        userid   = machine.RemoteUserid
        authdata = machine.AuthData
        host     = machine.LoginHost
        method   = self.AccessMethodByName(machine.AccessMethod)
        lcommand = command
        if startdir:
            CmdSep   = OSTypes[machine.OSType]["CommandSeparator"]
            lcommand = machine.GetChangeDirCommand(startdir) + CmdSep + lcommand
        result, output = method.RemoteExec(userid, host, authdata, lcommand, machine, self)
        if not (machine.StoreAuthData or store_auth):
            machine.AuthData = ""
        return (result, output)
    def MungeStringToPath (self, text):
        result = text.replace(" ", "_")
        result = result.replace("	", "_")
        result = result.replace("/", "-")
        result = result.replace("\\", "-")
        result = result.replace("(", "-")
        result = result.replace(")", "-")
        result = result.replace("*", "=")
        result = result.replace(";", ":")
        result = result.replace(",", "-")
        return result
    def RemotePath (self, pathlist, machname, absolute=False):
        machine = self.Machines.EntryByName(machname)
        result = ""
        if machine:
            PathSep = OSTypes[machine.OSType]["PathSeparator"]
            if absolute:
                result = PathSep
            nitems = len(pathlist)
            for i in range(nitems):
                if (pathlist[i] != "."):
                    if (i < nitems-1):
                        result = result + pathlist[i] + PathSep
                    else:
                        result = result + pathlist[i]
        return result
    def PathList (self, path):
        dir, name = os.path.split(path)
        if dir and (dir != "/") and (dir != "\\"):
            result = self.PathList(dir)
            result.append(name)
            return result
        else:
            return [name]
    def RemotePathList (self, path, machname):
        result = []
        machine = self.Machines.EntryByName(machname)
        print "  RemotePathList:  machine = ", machine, " (name ", machname, ')'
        if machine:
            PathSep = OSTypes[machine.OSType]["PathSeparator"]
            result = path.split(PathSep)
            print "               PathSep: ", PathSep
            print "               result:  ", result
        return result
    def MD5Diff (self, filelist1, filelist2):
        """
        MD5 differencing method based on Python Cookbook recipe by Stefano Spinucci at
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/304095.  This version
        only returns the list of files present in filelist1 but not in filelist2 and the
        list of files present in both lists for which the MD5's differ.
        """
        list_tmp = [line for line in filelist1.split("\n")]
        list_one = []    # initializing the first list
        for x in list_tmp:      # writing the first list
            if x[0:1] <> '#' and x[0:1] <> ';' and x.strip() <> '' :   # skip comment lines (starting with '#' or ';') and empty lines
                str_md5 = x[:32]    # extract the md5
                str_path = string.replace(os.path.dirname(x[33:]),'\\','/')    # extract the path
                if str_path[0:1] == '/' : str_path = str_path [1:]    # remove the trailing '/' from the path
                str_filename = os.path.basename(x[33:]).lstrip()    # extract the filename
                list_one.append((str_md5, str_path, str_filename))

        list_tmp = [line for line in filelist2.split("\n")]
        list_two = []    # initializing the first list
        for x in list_tmp:      # writing the first list
            if x[0:1] <> '#' and x[0:1] <> ';' and x.strip() <> '' :   # skip comment lines (starting with '#' or ';') and empty lines
                str_md5 = x[:32]    # extract the md5
                str_path = string.replace(os.path.dirname(x[33:]),'\\','/')    # extract the path
                if str_path[0:1] == '/' : str_path = str_path [1:]    # remove the trailing '/' from the path
                str_filename = os.path.basename(x[33:]).lstrip()    # extract the filename
                list_two.append((str_md5, str_path, str_filename))

        list_tmp = []
        # creating the two lists containing the result of the diff'ing
        list_one_diff = []
        list_two_diff = []
	# searching for '==' equal (md5 =, path =, filename =)
	for x in range(len(list_one)):
	    item_one = (list_one[x][0], list_one[x][1],list_one[x][2])
	    for y in range(len(list_two)):
	        item_two = (list_two[y][0], list_two[y][1],list_two[y][2])
	        if item_one == item_two:
	            list_one_diff.append([list_one[x][0], '==', list_one[x][1], list_one[x][2]])   # write in the first diff'ing result list
	            list_two_diff.append([list_two[y][0], '==', list_two[y][1], list_two[y][2]])   # write in the second diff'ing result list
	            list_one[x]=[]   # mark the first list element for removing
	            del list_two[y]   # remove the current item from the second list
	            break   # return to the upper for loop
	
	list_one = [x for x in list_one if x <> []]   # remove marked items from the first list

        # searching for '<>' different (md5 <>, path =, filename =)
        for x in range(len(list_one)):
            item_one = (list_one[x][1],list_one[x][2])
            for y in range(len(list_two)):
                item_two = (list_two[y][1],list_two[y][2])
                if item_one == item_two:
                    list_one_diff.append([list_one[x][0], '<>', list_one[x][1], list_one[x][2]])   # write in the first diff'ing result list
                    list_two_diff.append([list_two[y][0], '<>', list_two[y][1], list_two[y][2]])   # write in the second diff'ing result list
                    list_one[x]=[]   # mark the first list element for removing
                    del list_two[y]   # remove the current item from the second list
                    break   # return to the upper for loop

        list_one = [x for x in list_one if x <> []]   # remove marked items from the first list

        # searching for '>>' new dx and '<<' new sx (md5 n/a, path <>, filename <>)
        for x in range(len(list_one)):
            list_one_diff.append([list_one[x][0], '<<', list_one[x][1], list_one[x][2]])   # write in the first diff'ing result list

        #for y in range(len(list_two)):
        #    list_two_diff.append([list_two[y][0], '>>', list_two[y][1], list_two[y][2]])   # write in the second diff'ing result list

        # return result in same format as os.walk, ie. [(root, dirs, files), ...]
    
        result = [ ]
        
        for x in range(len(list_one_diff)):   # loops on the first diff'ing list, and print each element
            if (list_one_diff[x][1] != "=="):
                item = (os.path.dirname(list_one_diff[x][2]), [list_one_diff[x][2]], [])
                if (result.count(item) == 0):
                    result.append(item)
                result.append((list_one_diff[x][2], [], [list_one_diff[x][3]]))
        
        return result
    def Finalize (self):
        if os.path.exists(TempFile):
            os.remove(TempFile)
    def ParseRangeString (self, textlist):
        result    = [ ]
        wordsplit = re.compile(r'(.*)(?<!\\),(.*)')
        intrange  = re.compile(r'([0-9]+)-([0-9]+)')
        for item in textlist:
            if (item != ""):
                if re.match(wordsplit, item):
                    result.extend(self.ParseRangeString(re.split(wordsplit, item)))
                elif re.match(intrange, item):
                    split    = re.split(intrange, item)
                    bvalues  = [int(split[1]), int(split[2])]
                    minvalue = min(bvalues)
                    maxvalue = max(bvalues)
                    values   = range(minvalue, maxvalue+1)
                    result.extend([str(v) for v in values])
                else:
                    result.append(item)
        return result
            
# Container class for the different types of projects, experiments, runs, and
# jobs manipulated by the user.
    
class Container:
    def __init__ (self, parent, creator, name, comment):
        self.Parent       = parent
        self.Children     = [ ]
        self.BeginDate    = time.asctime(time.localtime())
        self.ModifiedDate = time.asctime(time.localtime())
        self.Creator      = creator
        self.CreatorName  = creator.Name
        self.Name         = name
        self.Comment      = comment
    def AddChild (self):
        pass
    def Delete (self):
        # kill my children and notify my parent (if I have one) that I'm dead
        for Child in self.Children:
            Child.Delete()
        if self.Parent:
            self.Parent.Children.remove(self)
        del self
    def Backup (self):
        Copy        = copy.deepcopy(self)
        Copy.Parent = self.Parent
        Copy.SetParentLinks()
        return Copy
    def RestoreBackup (self, SavedCopy):
        self.__dict__ = SavedCopy.__dict__
        self.SetParentLinks()
        del SavedCopy
    def Clone (self, user):
        Copy              = copy.deepcopy(self)
        Copy.Name         = Copy.Name + " Copy"
        Copy.BeginDate    = time.asctime(time.localtime())
        Copy.ModifiedDate = time.asctime(time.localtime())
        Copy.Creator      = user
        Copy.CreatorName  = user.Name
        Copy.Parent       = self.Parent
        if Copy.Parent:
            Copy.Parent.Children.append(Copy)
        Copy.SetParentLinks()
        return Copy
    def SetParentLinks (self):
        for Child in self.Children:
            Child.Parent = self
            Child.SetParentLinks()
    def Print (self, indent=""):
        print "%s%s %s (parent: %s)" % (indent, self.Name, self, self.Parent)
        for Child in self.Children:
            Child.Print(indent+"    ")
    def StatusLine (self):
        return ""
    
# Project class.

class Project(Container):
    def __init__ (self, user):
        Container.__init__(self, None, user, "New Project", "")
        self.NotesLocation = ""
        self.NotesType     = SM_WIKI
    def AddChild (self, user):
        NewExperiment = Experiment(user, self)
        # if we didn't press cancel...
        self.Children.append(NewExperiment)
        return NewExperiment

# Experiment class.

class Experiment(Container):
    def __init__ (self, user, parent):
        Container.__init__(self, parent, user, "New Experiment", "")
        self.Application     = ""
        self.ConfigFile      = ""
        self.RemoteBuildDir  = ""
        self.ConfigCommand   = ""
        self.BuildCommand    = ""
        self.Executable      = ""
        self.ExecArgs        = ""
        self.DataSrcMachine  = ""
        self.DataSrcFiles    = ""
        self.DataDestMachine = ""
        self.DataDestPath    = ""
        self.ExecMachine     = ""
        self.ExecQueue       = ""
        self.ExecAccount     = ""
        self.NumCPUsRange    = "1"
        self.Tiling          = 1
        self.MemoryPerNode   = 1000
        self.ParmFileTemplate= ""
        self.ParmName1       = ""
        self.ParmRange1      = ""
        self.ParmLock1       = 0
        self.ParmName2       = ""
        self.ParmRange2      = ""
        self.ParmLock2       = 0
        self.ParmName3       = ""
        self.ParmRange3      = ""
        self.ParmLock3       = 0
        self.ParmName4       = ""
        self.ParmRange4      = ""
        self.ParmLock4       = 0
        self.ParmName5       = ""
        self.ParmRange5      = ""
        self.ParmLock5       = 0
        self.ParmName6       = ""
        self.ParmRange6      = ""
        self.ParmLock6       = 0
        self.ParmName7       = ""
        self.ParmRange7      = ""
        self.ParmLock7       = 0
        self.ParmName8       = ""
        self.ParmRange8      = ""
        self.ParmLock8       = 0
        self.ConfigOutput    = ""
        self.ConfigStatus    = SM_UNKNOWN
        self.BuildOutput     = ""
        self.BuildStatus     = SM_UNKNOWN
    def AddChild (self, user, name, values, Npstr):
        if self.ParmFileTemplate:
            try:
                f = open(self.ParmFileTemplate)
                template = f.read()
                f.close()
                NewRun = Run(user, self)
                NewRun.Name         = name
                NewRun.NumCPUs      = int(Npstr)
                NewRun.ParmFileData = template
                for parm in values.keys():
                    NewRun.ParmVariables = NewRun.ParmVariables + \
                                           parm + " = " + values[parm] + "\n"
                    # use re.sub() to match patterns like ParmName = %ParmName%, replace
                    # matched expressions with given values, place result in ParmFileData
                    pattern = re.compile(r'%' + parm + r'%')
                    NewRun.ParmFileData = re.sub(pattern, values[parm], NewRun.ParmFileData)
                # if we didn't press cancel...
                self.Children.append(NewRun)
                return NewRun
            except IOError:
                return None
        else:
            NewRun = Run(user, self)
            NewRun.Name    = name
            NewRun.NumCPUs = int(Npstr)
            self.Children.append(NewRun)
            return NewRun
    def AddChildren (self, store):
        NewChildren   = [ ]
        self.ParseParms(store)
        values        = { }
        indexes       = { }
        counts        = { }
        Nruns         = 1
        locks_counted = False
        imin          = len(self.Children)
        for parm in self.Parms.keys():
            values[parm]  = ""
            indexes[parm] = 0
            counts[parm]  = len(self.Parms[parm])
            if (not self.Locks[parm]) or (self.Locks[parm] and (not locks_counted)):
                Nruns = Nruns * counts[parm]
                if self.Locks[parm]:
                    locks_counted = True
                    Nlockvalues   = counts[parm]
        for Npstr in self.NumCPUs:
            for parm in indexes.keys():
                indexes[parm] = 0
            for i in xrange(imin, Nruns+imin):
                for parm in indexes.keys():
                    values[parm] = self.Parms[parm][indexes[parm]]
                k = i + 1
                child_name = ""
                while k > 0:
                    child_name = child_name + chr(((k-1) % 26)+ord("A"))
                    k = (k-1) / 26
                L = list(child_name)
                L.reverse()
                child_name = ""
                for char in L:
                    child_name = child_name + char
                if (len(self.NumCPUs) > 1):
                    child_name = child_name + Npstr
                Child = self.AddChild(store.CurrentUser, child_name, values, Npstr)
                NewChildren.append(Child)
                j = 0
                locks_incremented = False
                while j < len(indexes.keys()):
                    parm = indexes.keys()[j]
                    if self.Locks[parm]:
                        if locks_incremented:
                            j = j + 1
                        else:
                            for lparm in indexes.keys():
                                if self.Locks[lparm]:
                                    indexes[lparm] = indexes[lparm] + 1
                            locks_incremented = True
                    else:
                        indexes[parm] = indexes[parm] + 1
                    if (indexes[parm] == counts[parm]):
                        if self.Locks[parm]:
                            for lparm in indexes.keys():
                                if self.Locks[lparm]:
                                    indexes[lparm] = 0
                        else:
                            indexes[parm] = 0
                        j = j + 1
                    else:
                        break
        return NewChildren
    def ParseParms (self, store):
        self.Parms = { }
        self.Locks = { }
        if (self.ParmName1 != ""):
            self.Parms[self.ParmName1] = store.ParseRangeString([self.ParmRange1])
            self.Locks[self.ParmName1] = self.ParmLock1
        if (self.ParmName2 != ""):
            self.Parms[self.ParmName2] = store.ParseRangeString([self.ParmRange2])
            self.Locks[self.ParmName2] = self.ParmLock2
        if (self.ParmName3 != ""):
            self.Parms[self.ParmName3] = store.ParseRangeString([self.ParmRange3])
            self.Locks[self.ParmName3] = self.ParmLock3
        if (self.ParmName4 != ""):
            self.Parms[self.ParmName4] = store.ParseRangeString([self.ParmRange4])
            self.Locks[self.ParmName4] = self.ParmLock4
        if (self.ParmName5 != ""):
            self.Parms[self.ParmName5] = store.ParseRangeString([self.ParmRange5])
            self.Locks[self.ParmName5] = self.ParmLock5
        if (self.ParmName6 != ""):
            self.Parms[self.ParmName6] = store.ParseRangeString([self.ParmRange6])
            self.Locks[self.ParmName6] = self.ParmLock6
        if (self.ParmName7 != ""):
            self.Parms[self.ParmName7] = store.ParseRangeString([self.ParmRange7])
            self.Locks[self.ParmName7] = self.ParmLock7
        if (self.ParmName8 != ""):
            self.Parms[self.ParmName8] = store.ParseRangeString([self.ParmRange8])
            self.Locks[self.ParmName8] = self.ParmLock8
        if (self.NumCPUsRange == ""):
            self.NumCPUsRange = "1"
        self.NumCPUs = store.ParseRangeString([self.NumCPUsRange])
    def CheckData (self, store):
        # check to make sure parameter & CPU ranges, etc. are OK
        self.ParseParms(store)
        result    = SM_OK
        lockcount = 0
        for parm in self.Parms.keys():
            if (len(self.Parms[parm]) == 0):
                result = SM_FAIL
                break
            if self.Locks[parm]:
                if (lockcount == 0):
                    lockcount = len(self.Parms[parm])
                else:
                    if (len(self.Parms[parm]) != lockcount):
                        result = SM_FAIL
                        break
        return result
    

# Run class.

class Run(Container):
    def __init__ (self, user, parent):
        Container.__init__(self, parent, user, "New Run", "")
        self.Disposition     = ""
        self.Application     = parent.Application
        self.Executable      = parent.Executable
        self.ExecArgs        = parent.ExecArgs
        self.ExecMachine     = parent.ExecMachine
        self.Queue           = parent.ExecQueue
        self.Account         = parent.ExecAccount
        self.NumCPUs         = 1
        self.Tiling          = parent.Tiling
        self.MemoryPerNode   = parent.MemoryPerNode
        self.DataSrcMachine  = parent.DataSrcMachine
        self.DataSrcFiles    = parent.DataSrcFiles
        self.DataDestMachine = parent.DataDestMachine
        self.DataDestPath    = parent.DataDestPath
        self.ParmFileData    = ""
        self.ParmTemplate    = parent.ParmFileTemplate
        self.ParmVariables   = ""
        self.JobCounter      = 0
    def AddChild (self, user):
        self.JobCounter = self.JobCounter + 1
        job_name = "%-0.3s%04u" % (self.Name.upper(), self.JobCounter)
        NewJob = Job(user, self, job_name)
        # if we didn't press cancel...
        self.Children.append(NewJob)
        return NewJob
    def AddChildren (self):
        # generate runs
        pass
    def StatusLine (self):
        return self.Disposition

# Job class.

class Job(Container):
    def __init__ (self, user, parent, name, type="Original"):
        Container.__init__(self, parent, user, name, "")
        self.Disposition     = "No data"
        self.Application     = parent.Application
        self.Executable      = parent.Executable
        self.ExecArgs        = parent.ExecArgs
        self.ExecMachine     = parent.ExecMachine
        self.Queue           = parent.Queue
        self.Account         = parent.Account
        self.NumCPUs         = parent.NumCPUs
        self.Tiling          = parent.Tiling
        self.MemoryPerNode   = parent.MemoryPerNode
        self.DataSrcMachine  = parent.DataSrcMachine
        self.DataSrcFiles    = parent.DataSrcFiles
        self.DataDestMachine = parent.DataDestMachine
        self.DataDestPath    = parent.DataDestPath
        self.ParmFileData    = parent.ParmFileData
        self.RemoteJobID     = "000000"
        self.SubmitDate      = "Not yet submitted"
        self.RunStatus       = SM_STATUS_NOTAVAIL
        self.DataStatus      = SM_STATUS_NOTAVAIL
        self.LogFileData     = ""
        self.StdoutData      = ""
        self.StderrData      = ""
        self.RemoteScriptFile= RemoteScriptFile
        self.RemoteStdoutFile= "000000.out"
        self.RemoteStderrFile= "000000.err"
        self.JobScript       = ""
        self.WallTimeHrs     = "00"
        self.WallTimeMin     = "00"
        self.WallTimeSec     = "00"
        self.RemoteRunDir    = ""
        self.WallTime        = "00:00"
        self.Comment         = type
        self.RemoteJobName   = ""
        self.LogFileFinal    = False
        self.StdoutFinal     = False
        self.StderrFinal     = False
    def Clone (self, user):
        clone = Container.Clone(self, user)
        clone.Disposition     = "No data"
        clone.RemoteJobID     = "000000"
        clone.SubmitDate      = "Not submitted yet"
        clone.RunStatus       = SM_STATUS_NOTAVAIL
        clone.DataStatus      = SM_STATUS_NOTAVAIL
        clone.LogFileData     = ""
        clone.StdoutData      = ""
        clone.StderrData      = ""
        clone.LogFileFinal    = False
        clone.StdoutFinal     = False
        clone.StderrFinal     = False
        return clone
    def GenerateJobScript (self, store):
        machine = store.Machines.EntryByName(self.ExecMachine)
        if (machine.JobTemplateFile != ""):
            try:
                f = open(machine.JobTemplateFile)
                template = f.read()
                f.close()
            except IOError:
                return SM_FAIL
        else:
            system = store.QueuingSystemByName(machine.QueueSystem)
            if system:
                template = system.JobTemplate
            else:
                template = self.Executable + " " + self.ExecArgs
        Run        = self.Parent
        Experiment = Run.Parent
        Project    = Experiment.Parent
        self.RemoteRunDir = store.RemotePath([machine.RemoteRunRootDir, \
                                              store.MungeStringToPath(Project.Name), \
                                              store.MungeStringToPath(Experiment.Name), \
                                              store.MungeStringToPath(Run.Name), \
                                              store.MungeStringToPath(self.Name)], \
                                             machine.Name)
        try:
            hours = int(self.WallTimeHrs)
        except ValueError:
            hours = 0
        hours = "%02u" % hours
        try:
            minutes = int(self.WallTimeMin)
        except ValueError:
            minutes = 0
        minutes = "%02u" % minutes
        try:
            seconds = int(self.WallTimeSec)
        except ValueError:
            seconds = 0
        seconds = "%02u" % seconds
        wall_time = "%s:%s" % (hours, minutes)
        self.WallTime = wall_time
        num_nodes = self.NumCPUs / self.Tiling
        app = store.Applications.EntryByName(self.Application)
        if app.IsParallel:
            pattern = ParsableString(machine.ParallelExecCommand)
            par_exec_cmd = pattern.Parse({SM_NUM_CPUS  : str(self.NumCPUs), \
                                          SM_NUM_NODES : str(num_nodes), \
                                          SM_TILING    : str(self.Tiling)})
        else:
            par_exec_cmd = ""
        pattern = ParsableString(template)
        jobname = store.MungeStringToPath(self.Name)
        self.RemoteJobName = jobname
        self.JobScript = pattern.Parse({SM_NUM_CPUS          : str(self.NumCPUs), \
                                        SM_NUM_NODES         : str(num_nodes), \
                                        SM_MEM_PER_NODE      : str(self.MemoryPerNode), \
                                        SM_WALLTIME          : wall_time, \
                                        SM_WALLTIME_HOURS    : hours, \
                                        SM_WALLTIME_MINS     : minutes, \
                                        SM_WALLTIME_SECS     : seconds, \
                                        SM_TILING            : str(self.Tiling), \
                                        SM_QUEUE             : self.Queue, \
                                        SM_ACCOUNT           : self.Account, \
                                        SM_JOB_NAME          : jobname, \
                                        SM_PATH_CREATE       : machine.GetMakeDirCommand(), \
                                        SM_CHDIR             : machine.GetChangeDirCommand(), \
                                        SM_PAR_EXEC          : par_exec_cmd, \
                                        SM_REMOTE_EXEC       : self.Executable, \
                                        SM_ARGS              : self.ExecArgs, \
                                        SM_REMOTE_RUN_PATH   : self.RemoteRunDir, \
                                        SM_REMOTE_TEUTHISDIR : machine.TeuthisDir})
        return SM_OK
    def StatusLine (self):
        if not hasattr(self, "WallTime"):
            self.WallTime = ""
        if (self.NumCPUs == 1):
            CPUstring = "CPU"
        else:
            CPUstring = "CPUs"
        try:
            Date = time.strftime("%H:%M %m/%d/%Y", time.strptime(self.SubmitDate))
        except ValueError:
            Date = self.SubmitDate
        return self.RemoteJobID + " [" + Date + "] " + \
               str(self.NumCPUs) + " " + CPUstring + "/" + self.WallTime + \
               " (" + self.RunStatus + ") " + self.Disposition
    def CreateRestartJob (self, store, parms=""):
        clone         = self.Clone(store.CurrentUser)
        clone.Name    = self.Name
        clone.Comment = "Restart of " + self.RemoteJobID
        app = store.Applications.EntryByName(self.Application)
        if (app.RestartType == SM_RESTART_ARGS):
            clone.ExecArgs = app.RestartArgs + clone.ExecArgs
        elif (app.RestartType == SM_RESTART_COPY):
            clone.ParmFileData = parms
        else: # SM_RESTART_CONTROL
            pass   # check for existence of app.RestartControlFile or something
        return clone

# Class to contain user data.

class User:
    def __init__ (self):
        self.Name                     = ""
        self.Organization             = ""
        self.Email                    = ""
        self.Workspace                = HomeDirectory
        self.DataEntered              = False
        self.ProjectTextColor         = (0,0,0)
        self.ExperimentTextColor      = (0,0,0)
        self.RunTextColor             = (0,0,0)
        self.JobTextColor             = (0,0,0)
        self.ProjectTextFont          = "Sans 10"
        self.ExperimentTextFont       = "Sans 10"
        self.RunTextFont              = "Sans 10"
        self.JobTextFont              = "Sans 10"
        self.TreeViewShowDescription  = True
        self.TreeViewShowStatus       = True
        self.TreeViewShowModifiedDate = True
        self.ConfirmDeletions         = True
        self.ShowOutputOnError        = False
        self.UseTendril               = False
        self.VerboseLogging           = False
        self.LogFile                  = "teuthis.log"
        self.GridCertDir              = UserGlobusDirectory
        self.GridProxyLifetime        = 12
        self.UseMyProxy               = False
        self.MyProxyServer            = ""
        self.MyProxyAccount           = ""
        self.AutoUpdate               = False
        self.AutoUpdateInterval       = 10

# Resource list class for applications, machines, etc.
        
class ResourceList:
    def __init__ (self, store):
        self.Entries   = [ ]
        self.Current   = None
        self.New(store)
    def New (self, store):
        self.Current = Resource(store)
        self.Add()
    def Add (self):
        if not (self.Current in self.Entries):
            self.Entries.append(self.Current)
    def Extend (self, reslist):
        self.Entries.extend(reslist.Entries)
    def Clone (self):
        if (self.Current in self.Entries):
            Copy = copy.deepcopy(self.Current)
            self.Current = Copy
            self.Current.Name = self.Current.Name + " Copy"
            self.Add()
    def Delete (self, store):
        self.Entries.remove(self.Current)
        self.Current.Delete()
        if self.Entries:
            self.Current = self.Entries[len(self.Entries)-1]
        else:
            self.New(store)
    def EntryByName (self, name):
        object = None
        for Entry in self.Entries:
            if (Entry.Name == name):
                object = Entry
        return object
    def EntryNames (self):
        List = [ ]
        for Entry in self.Entries:
            List.append(Entry.Name)
        return List
    def Print (self):
        pass

# Class to handle application lists

class ApplicationList(ResourceList):
    def New (self, store):
        self.Current = Application(store)
        self.Add()

# Class to handle machine lists

class MachineList(ResourceList):
    def New (self, store):
        self.Current = Machine(store)
        self.Add()
    def Clone (self):
        ResourceList.Clone(self)
        self.Current.AlreadyPrompted = False
        self.Current.StoreAuthData   = False
        self.Current.AuthData        = ""
        
# Class to handle individual resources like applications and machines

class Resource:
    def __init__ (self, store):
        pass
    def Delete (self):
        del self
    def Print (self):
        pass

# Class to handle application data

class Application(Resource):
    def __init__ (self, store):
        self.Name                = "New Application"
        self.Description         = ""
        self.LocalSourcePath     = ""
        self.ExecutableName      = ""
        self.ConfigFileName      = ""
        self.ConfigCommand       = ""
        self.BuildCommand        = ""
        self.ParamFileName       = ""
        self.LogFileName         = ""
        self.BuildExecutable     = False
        self.MoveExecutable      = False
        self.IsParallel          = False
        self.RestartType         = SM_RESTART_CONTROL
        self.RestartControlFile  = ""
        self.RestartArgs         = ""
        self.RestartCopyFromFile = ""
        Resource.__init__(self, store)

# Class to handle machine data

class Machine(Resource):
    def __init__ (self, store):
        self.Name                = "New Machine"
        self.Description         = ""
        self.Type                = SM_COMPUTE_MACHINE
        self.LoginHost           = ""
        self.OSType              = "Unix"
        self.AccessMethod        = store.LocalMethod.Name
        self.AccessMethodType    = store.LocalMethod.Type
        self.Realm               = ""
        self.GridCertSubject     = ""
        self.GridCertFile        = ""
        self.GridKeyFile         = ""
        self.QueueSystem         = store.UnixNoQueue.Name
        self.AuthInitCommand     = store.LocalMethod.AuthInitCommand.String
        self.RemoteExecCommand   = store.LocalMethod.RemoteExecCommand.String
        self.FileUploadCommand   = store.LocalMethod.FileUploadCommand.String
        self.FileDownloadCommand = store.LocalMethod.FileDownloadCommand.String
        self.ThirdPartyTransferCmd = ""
        self.JobTemplateFile     = ""
        self.CPUsPerNode         = 1
        self.DefaultCPUsPerNode  = 1
        self.MemoryPerNode       = 1000
        self.RemoteUserid        = ""
        self.DefaultQueue        = ""
        self.DefaultAccount      = ""
        self.RemoteBuildRootDir  = ""
        self.RemoteExecDir       = ""
        self.RemoteRunRootDir    = ""
        self.TeuthisDir          = ""
#        self.FileCopyCommand     = "cp"
#        self.FileMoveCommand     = "mv"
#        self.PathCreateCommand   = "mkdir -p"
#        self.ChangeDirCommand    = "cd"
        self.ParallelExecCommand = ""
        self.Queues              = [ ]
        self.Accounts            = [ ]
        self.AuthDataName        = ""
        self.AuthData            = ""
        self.StoreAuthData       = False
        self.AlreadyPrompted     = False
        Resource.__init__(self, store)
    def GetMakeDirCommand (self, path=""):
        if (self.OSType == "Unix"):
            return "mkdir -p " + path
        elif (self.OSType == "Windows"):
            return "mkdir " + path
        else:
            return "ERROR"
    def GetChangeDirCommand (self, path=""):
        if (self.OSType == "Unix"):
            return "cd " + path
        elif (self.OSType == "Windows"):
            return "chdir " + path
        else:
            return "ERROR"
    def GetMakeExecutableCommand (self, path=""):
        if (self.OSType == "Unix"):
            return "chmod u+x " + path
        else:
            return "ERROR"
    def GetFileCopyCommand (self, src_files="", dest=""):
        if (self.OSType == "Unix"):
            return "cp " + src_files + " " + dest
        elif (self.OSType == "Windows"):
            return "copy " + src_files + " " + dest
        else:
            return "ERROR"
    def GetFileMoveCommand (self, src_files="", dest=""):
        if (self.OSType == "Unix"):
            return "mv " + src_files + " " + dest
        elif (self.OSType == "Windows"):
            return "move " + src_files + " " + dest
        else:
            return "ERROR"

# Class to handle strings containing keywords that we parse

class ParsableString:
    def __init__ (self, text):
        self.String = text
    def Parse (self, dict):
        result = self.String
        for token in dict.keys():
            result = result.replace(token, dict[token])
        return result

# Class to handle machine access methods

class AccessMethod:
    def __init__ (self, Name, Init, RExec, Upload, Download, Type, PromptName, PromptData, Store, \
                  ThirdPartyTransfer=None):
        self.Name                = Name
        self.AuthInitCommand     = ParsableString(Init)
        self.RemoteExecCommand   = ParsableString(RExec)
        self.FileUploadCommand   = ParsableString(Upload)
        self.FileDownloadCommand = ParsableString(Download)
        if ThirdPartyTransfer:
            self.ThirdPartyTransferCmd = ThirdPartyTransfer
        else:
            self.ThirdPartyTransferCmd = ""
        self.Type                = Type
        self.AuthDataName        = PromptName
        self.AuthDataPrompt      = PromptData
        self.StoreAuthData       = Store
    def Exec (self, command, dict, password, store, ForcePrompt=False):
        cmd = command.Parse(dict)
        if not cmd:
            return (SM_FAIL, "")
        output = ""
        status = SM_OK
        store.LogMessage("Execute: %s" % cmd)
        if (self.Type == SM_AUTH_PASSWORD) or ForcePrompt:
            try:
                process = pexpect.spawn(cmd)
            except pexpect.ExceptionPexpect, e:
                return (SM_FAIL, "%s" % e)
            store.LiveRemoteProcess = process
            # ugly, somewhat ssh-specific.
            index = process.expect([self.AuthDataPrompt, pexpect.EOF, \
                                    "continue connecting (yes/no)?"])
            output = output + "%s%s" % (process.before, process.after)
            if (index == 2):
                process.sendline("yes")
                index = process.expect([self.AuthDataPrompt, pexpect.EOF])
                output = output + "%s%s" % (process.before, process.after)
            if (index == 0):
                process.sendline(password)
                index = process.expect([self.AuthDataPrompt, pexpect.EOF], timeout=SM_TIMEOUT)
                output = output + "%s%s" % (process.before, process.after)
                if (index == 0):
                    status = SM_FAIL
            else:
                status = SM_FAIL
            if ((status == SM_OK) and (index != 1)):
                process.expect(pexpect.EOF, timeout=SM_TIMEOUT)
                output = output + "%s%s" % (process.before, process.after)
            process.close(wait=1)
            if process.exitstatus:
                status = SM_FAIL
        else:
            try:
                process = pexpect.spawn(cmd)
            except pexpect.ExceptionPexpect, e:
                return (SM_FAIL, "%s" % e)
            store.LiveRemoteProcess = process
            index = process.expect([pexpect.EOF, pexpect.TIMEOUT], timeout=SM_TIMEOUT)
#            success = False
#            ntries  = 0
#            while (not success) and (ntries < 100):
#                try:
#                    index = process.expect([pexpect.EOF, pexpect.TIMEOUT], timeout=SM_TIMEOUT)
#                    success = True
#                except select.error:
#                    ntries += 1
#                    if (ntries == 100): index = 1
            output = output + "%s%s" % (process.before, process.after)
            process.close(wait=1)
            if process.exitstatus or (index == 1):
                status = SM_FAIL
            else:
                status = SM_OK
        store.LiveRemoteProcess = None
        return (status, output)
    def GetCredentials (self, user, host, realm, password, hostobj, store):
        status, output = self.Exec(ParsableString(hostobj.AuthInitCommand), \
                                        { SM_REMOTE_USER : user, \
                                          SM_REMOTE_HOST : host, \
                                          SM_REMOTE_PASS : password, \
                                          SM_REMOTE_REALM: realm, \
                                          SM_CERT_FILE   : hostobj.GridCertFile, \
                                          SM_KEY_FILE    : hostobj.GridKeyFile,
                                          SM_PROXY_LIFE  : str(store.CurrentUser.GridProxyLifetime)}, \
                                        password, store, ForcePrompt=True )
        return (status, output)
    def RemoteExec (self, user, host, password, command, hostobj, store):
        status, output = self.Exec(ParsableString(hostobj.RemoteExecCommand), \
                                          { SM_REMOTE_USER : user, \
                                            SM_REMOTE_HOST : host, \
                                            SM_REMOTE_PASS : password, \
                                            SM_REMOTE_CMD  : command }, \
                                          password, store )
        return (status, output)
    def UploadFile (self, user, host, password, lpath, rpath, hostobj, store):
        status, output = self.Exec(ParsableString(hostobj.FileUploadCommand), \
                                          { SM_REMOTE_USER : user, \
                                            SM_REMOTE_HOST : host, \
                                            SM_REMOTE_PASS : password, \
                                            SM_LOCAL_PATH  : lpath, \
                                            SM_REMOTE_PATH : rpath }, \
                                          password, store )
        return (status, output)
    def DownloadFile (self, user, host, password, lpath, rpath, hostobj, store):
        status, output = self.Exec(ParsableString(hostobj.FileDownloadCommand), \
                                            { SM_REMOTE_USER : user, \
                                              SM_REMOTE_HOST : host, \
                                              SM_REMOTE_PASS : password, \
                                              SM_LOCAL_PATH  : lpath, \
                                              SM_REMOTE_PATH : rpath }, \
                                            password, store )
        return (status, output)
    def TransferThirdPartyFile (self, user1, user2, password, host1, host2, path1, path2, host1obj, host2obj, store):
        if (not host1obj.ThirdPartyTransferCmd) or (not host2obj.ThirdPartyTransferCmd) or \
           (host1obj.ThirdPartyTransferCmd != host2obj.ThirdPartyTransferCmd):
            return (SM_FAIL, "")
        else:
            status, output = self.Exec(ParsableString(host1obj.ThirdPartyTransferCmd), \
                                            { SM_REMOTE_USER : user1, \
                                              SM_REMOTE_USER2: user2, \
                                              SM_REMOTE_HOST : host1, \
                                              SM_REMOTE_HOST2: host2, \
                                              SM_REMOTE_PASS : password, \
                                              SM_LOCAL_PATH  : path1, \
                                              SM_REMOTE_PATH : path2 }, \
                                            password, store )
        return (status, output)

# Class to store information about queuing systems

class QueuingSystem:
    def __init__ (self, Name, SubmitCommand, StatusCommand, KillCommand, PeekCommand, JobTemplate, \
                  SubmitOutput, StdoutFile, StderrFile, StatusOutput, StatusDict):
        self.Name                 = Name
        self.SubmitCommand        = SubmitCommand
        self.StatusCommand        = StatusCommand
        self.KillCommand          = KillCommand
        self.PeekCommand          = PeekCommand
        self.JobTemplate          = JobTemplate
        self.SubmitOutput         = SubmitOutput
        self.RemoteStdoutFileName = StdoutFile
        self.RemoteStderrFileName = StderrFile
        self.StatusOutput         = StatusOutput
        self.StatusDict           = StatusDict
    def GetSubmitCommand (self, job):
        return ParsableString(self.SubmitCommand).Parse({SM_BATCH_FILE : RemoteScriptFile})
    def GetPeekCommand (self, job):
        return ParsableString(self.PeekCommand).Parse({SM_REMOTE_JID : job.RemoteJobID})
    def GetStatusCommand (self, job):
        return ParsableString(self.StatusCommand).Parse({SM_REMOTE_JID : job.RemoteJobID})
    def GetRemoteStdoutFileName (self, job, machine):
        return ParsableString(self.RemoteStdoutFileName).Parse({SM_REMOTE_JID        : job.RemoteJobID, \
                                                                SM_JOB_NAME          : job.RemoteJobName, \
                                                                SM_REMOTE_RUN_PATH   : job.RemoteRunDir, \
                                                                SM_REMOTE_TEUTHISDIR : machine.TeuthisDir})
    def GetRemoteStderrFileName (self, job, machine):
        return ParsableString(self.RemoteStderrFileName).Parse({SM_REMOTE_JID        : job.RemoteJobID, \
                                                                SM_JOB_NAME          : job.RemoteJobName, \
                                                                SM_REMOTE_RUN_PATH   : job.RemoteRunDir, \
                                                                SM_REMOTE_TEUTHISDIR : machine.TeuthisDir})
    def GetJobID (self, text):
        if text:
            pattern = re.compile(self.SubmitOutput)
            result = pattern.findall(text)
            if (result != []):
                return result[0]
        return "000000"
    def GetJobStatus (self, text):
        if text:
            pattern = re.compile(self.StatusOutput)
            result = pattern.findall(text)
            if (result != []):
                result = result[0]
                if result in self.StatusDict.keys():
                    return self.StatusDict[result]
        return SM_STATUS_UNK

# Main event loop
        
def main():
    gtk.main()
 
if __name__ == "__main__":
    store        = AppStore()
    SM_LOCALHOST = Machine(store)
    app          = AppGUI(store)
    main()

